from six import text_type
from xml.sax.saxutils import quoteattr

from hypertag.core.errors import VoidTagEx, TypeErrorEx
from hypertag.core.dom import DOM


########################################################################################################################################################
#####
#####  TAG
#####

class Tag:
    """
    Base class for custom tags defined as a Python functions (external tags).
    Instances can be imported to a Hypertag script and used as tags.
    """
    name = None         # [str] name that indentifies this tag in a DOM and can be used in DOM selectors
    
    pure = True         # if True, the tag is assumed to always return the same result for the same arguments (no side effects),
                        # which enables caching of expand() calls and compactification of DOM nodes tagged with this tag
    
    def expand(self, body, attrs, kwattrs):
        """
        Subclasses should NOT append trailing \n nor add extra indentation during tag expansion
        - both things are added by the parser later on, if desired so by programmer.
        
        :param body: DOM of a translated body of tag occurrence, possibly empty; if a tag is void and does not expect a body,
                     but a non-empty body was passed, the tag should raise VoidTagEx
        :param attrs: list/tuple of positional attributes
        :param kwattrs: dict of keyword attributes; attributes are guaranteed to have valid XML names,
                        but NOT necessarily valid Python names
        :return: a string
        """
        raise NotImplementedError

    def __call__(self, body_string = None, *attrs, **kwattrs):
        """
        Wrapper around expand() to let the tag be used like a function and to allow its direct rendering,
        without an intermediate DOM. This method can be used as a shortcut inside Hypertag expressions
        to skip DOM creation entirely.
        """
        body = DOM.Text(text_type(body_string)) if body_string is not None else DOM()
        return self.expand(body, attrs, kwattrs)
        

########################################################################################################################################################

class TagFunction(Tag):
    """A wrapper that creates a Tag instance from a function. The `body` attribute is passed as a string to the function."""
    
    def __init__(self, fun, dom = False):
        self.fun  = fun
        self.name = fun.__name__
        self.dom  = dom
    
    def expand(self, body, attrs, kwattrs):
        if not self.dom: body = body.render()
        return self.fun(body, *attrs, **kwattrs)
        

class Null(Tag):
    """Null tag '.' is represented in the DOM tree. Its expand() passes the body unchanged."""
    
    name = 'null'
    
    def expand(self, body, attrs, kwattrs):
        return body.render()
    
null = Null()


########################################################################################################################################################
#####
#####  STANDARD MARKUP TAG
#####

class Markup(Tag):
    """
    A hypertag whose expand() outputs the body unchanged, surrounded by <name>...</name> strings, with proper handling
    of void tags <name /> and HTML/XHTML format differences for boolean attributes.
    This class is used for all built-in (X)HTML tags. It can also be used to define custom markup tags in an application.
    """
    
    name = None         # tag <name> to be printed into markup; may differ from the Hypertag name used inside a script (!)
    void = False        # True if this tag is void: the `body` in expand() should be empty and the element is rendered as self-closing: <NAME />
    mode = 'HTML'       # (X)HMTL compatibility mode: either 'HTML' or 'XHTML'
    
    def __init__(self, name = None, void = False, mode = 'HTML'):
        if name: self.name = name
        self.void = void
        self.mode = mode
    
    def expand(self, body, attrs, kwattrs):
        
        if attrs: raise TypeErrorEx("markup tag '%s' does not accept positional attributes" % self.name)
        
        name = self.name
        
        # render attributes
        kwattrs = filter(None, map(self._render_attr, kwattrs.items()))
        tag = ' '.join([name] + list(kwattrs))

        body = body.render()
        
        # render output
        if self.void:
            if body: raise VoidTagEx("non-empty body passed to a void markup tag '%s'" % name)
            return "<%s />" % tag
        else:
            # if the block contains a headline, the closing tag is placed on the same line as body;
            # a newline is added at the end, otherwise
            nl = '\n' if body[:1] == '\n' else ''
            return "<%s>" % tag + body + nl + "</%s>" % name

    def _render_attr(self, name_value):
        
        name, value = name_value
        if value is True:               # name=True   -- converted to:  name (HTML)  or  name="name" (XHTML)
            if self.mode == 'HTML':
                return name
            else:
                return '%s="%s"' % (name, name)
        if value is False:              # name=False  -- removed from attr list
            return None
        
        value = str(value)
        if '"' not in value:
            value = '"%s"' % value
        elif "'" not in value:
            value = "'%s'" % value
        else:
            value = quoteattr(value)    # escaping of <,>,&," chars is performed ONLY when the value contains a quote "
        
        return '%s=%s' % (name, value)
        




from hypertag.core.errors import TypeErrorEx
from hypertag.core.tag import Markup
from hypertag.std.registry import Registry


register = Registry()


########################################################################################################################################################
#####
#####  UTILITIES
#####

def html_escape(text):
    """Escape HTML/XML special characters (& < >) in a string that's to be embedded in a text part of an HTML/XML document.
    For escaping attribute values use html_attr_escape() - attributes need a different set of special characters to be escaped.
    """
    return text.replace('&', '&amp;').replace('>', '&gt;').replace('<', '&lt;')


########################################################################################################################################################
#####
#####  STANDARD HTML5 tags
#####
#####  In the future, we would like to provide a way to easily produce "polyglot" (X)HTML5 markup, which will be
#####  compatible with both HTML5 and XHTML5.
#####

_HTML_TAGS_VOID    = "area base br col command embed hr img input link meta param source track wbr".split()
_HTML_TAGS_NONVOID = "a abbr acronym address applet article aside audio b basefont bdi bdo big blockquote body " \
                     "button canvas caption center cite code colgroup data datalist dd del details dfn dialog dir " \
                     "div dl dt em fieldset figcaption figure font footer form frame frameset h1 h2 h3 h4 h5 h6 " \
                     "head header html i iframe ins kbd label legend li main map mark meter nav noframes noscript " \
                     "object ol optgroup option output p picture pre progress q rp rt ruby s samp script section " \
                     "select small span strike strong style sub summary sup svg table tbody td template textarea " \
                     "tfoot th thead time title tr tt u ul var video".split()

def _create(name, void):
    # register.tag(Markup(name.lower(), void, 'XHTML'))         # XHTML only permits lowercase tag names
    register.tag(Markup(name.lower(), void, 'HTML'))
    register.tag(Markup(name.upper(), void, 'HTML'))


### create tags

for tag in _HTML_TAGS_NONVOID:  _create(tag, False)
for tag in _HTML_TAGS_VOID:     _create(tag, True)
    

########################################################################################################################################################
#####
#####  HYPERTAG'S CUSTOM tags & functions (HTML-specific)
#####

@register.tag
class comment(Markup):
    
    name = "comment"
    
    def expand(self, body, attrs, kwattrs):
        
        if attrs or kwattrs: raise TypeErrorEx("'comment' tag does not accept attributes")
        body = body.render()
        
        # if the block contains a headline, the closing tag is placed on the same line as body;
        # a newline is added at the end, otherwise
        nl = '\n' if body[:1] == '\n' else ''
        return "<!--" + body + nl + "-->"
    

    


# -*- coding: utf-8 -*-
"""
@author:  Marcin Wojnarski
"""

from __future__ import unicode_literals

import sys, re, codecs, operator
from collections import OrderedDict
from six import reraise, text_type

from parsimonious.grammar import Grammar as Parsimonious
from parsimonious.exceptions import IncompleteParseError

from hypertag.nifty.util import ObjDict
from hypertag.nifty.parsing import ParsimoniousTree as BaseTree

from hypertag.core.errors import SyntaxErrorEx, ValueErrorEx, TypeErrorEx, FalseValueEx, NameErrorEx, \
    UnboundLocalEx, UndefinedTagEx, NotATagEx, NoneStringEx, VoidTagEx, ImportErrorEx
from hypertag.core.grammar import grammar, MARK_TAG, MARK_VAR, TAG, VAR, IS_TAG
from hypertag.core.xml_chars import XML_StartChar, XML_Char, XML_EndChar
from hypertag.core.structs import Context, State, Slot, ValueSlot
from hypertag.core.dom import del_indent, get_indent, DOM
from hypertag.core.tag import Tag, null

DEBUG = False


#####################################################################################################################################################
#####
#####  UTILITIES
#####

_re_decode_escapes = re.compile(r'''
    ( \\U........               # 8-digit hex escapes
    | \\u....                   # 4-digit hex escapes
    | \\x..                     # 2-digit hex escapes
    | \\[0-7]{1,3}              # Octal escapes
    | \\N\{[^}]+\}              # Unicode characters by name
    | \\[\\'"abfnrtv]           # Single-character escapes including "\\"
    )''', re.UNICODE | re.VERBOSE)

def decode_escapes(s):
    r"""
    Decoding of escape characters (\n \t \\ \x.. etc.) inside a string `s`. Handles Unicode correctly.
    Based on: https://stackoverflow.com/a/24519338/1202674
    """
    def decode_match(match):
        return codecs.decode(match.group(0), 'unicode-escape')
    
    return _re_decode_escapes.sub(decode_match, s)


def duplicate(seq):
    """Any duplicate in a sequence `seq`; or None if no duplicates are present."""
    seen = set()
    dups = set(x for x in seq if x in seen or seen.add(x))
    return dups.pop() if dups else None

def STR(value, node = None, msg = "expression to be embedded in markup text evaluates to None"):
    """Convert `value` to a string for embedding in text markup. Raise NoneStringEx if value=None."""
    if value is None: raise NoneStringEx(msg, node)
    return text_type(value)

def partial(func, *args, **kwargs):
    """
    Create a partial function during processing of a filter pipeline, like in x:fun(a,b)
    - the returned function represents application of (a,b) to a `fun`; whatever argument will be supplied later,
    it will be PREPENDED to the argument list, unlike in functools.partial() which APPENDS new arguments at the end.
    """
    def newfunc(*newargs):
        return func(*(newargs + args), **kwargs)
    return newfunc


#####################################################################################################################################################

class Native(Tag):
    """
    An already-expanded tag created by NODES.xblock_def for insertion into a DOM. Expansion of native tags is done
    before DOM creation, hence expand() - called after DOM creation - only performs body rendering and nothing else.
    """
    def __init__(self, name):
        self.name = name
        
    def expand(self, body, attrs, kwattrs):
        return body.render()


class Hypertag:
    """
    Base class for a hypertag definition block and for hypertags imported from other scripts.
    This class is UNRELATED to the Tag class: its expansion accepts more arguments and is performed
    BEFORE creation of a DOM, not after.
    """
    def expand(self, body, attrs, kwattrs, state, caller):
        raise NotImplementedError

class ImportedHypertag(Hypertag):
    """
    Wrapper around an imported hypertag. Stores the global state of the imported module,
    so that it can replace the state of the importing module when expand() of the imported hypertag is called.
    """
    def __init__(self, hypertag, module_symbols):
        
        assert isinstance(hypertag, Hypertag)
        self.hypertag = hypertag
        self.module_symbols = module_symbols
        
    def expand(self, body, attrs, kwattrs, state, caller):
        
        module_state = State(self.module_symbols)
        return self.hypertag.expand(body, attrs, kwattrs, module_state, caller)

class EmbeddedHypertag:
    """
    Wrapper around an imported or locally defined hypertag, for use inside expressions.
    Stores the current state of translation from the place of occurrence for subsequent expansion,
    plus the parent x* node of the AST for debugging.
    Allows the inner hypertag be called like a function through __call__():
    equivalent of Tag.__call__() for external tags.
    """
    def __init__(self, hypertag, state, caller):
        
        assert isinstance(hypertag, Hypertag)
        self.hypertag = hypertag
        self.caller = caller
        self.state  = state
        
    def __call__(self, body_string = None, *attrs, **kwattrs):
        
        body = DOM.Text(text_type(body_string)) if body_string is not None else DOM()
        dom  = self.hypertag.expand(body, attrs, kwattrs, self.state, self.caller)
        dom.set_indent('')
        return dom.render()


#####################################################################################################################################################
#####
#####  HYPERTAG GRAMMAR & SCRIPT PREPROCESSING
#####

class Grammar(Parsimonious):
    
    default = None      # class-level default instance of HyML_Parser, the one with standard indentation chars;
                        # can be used for parsing of a given text only if the text doesn't contain any of
                        # special characters that are used in the parser for indent / dedent
    
    SPECIAL_SYMBOLS = ['INDENT_S', 'DEDENT_S', 'INDENT_T', 'DEDENT_T']
    CHARS_DEFAULT   = [u'\u2768', u'\u2769', u'\u276A', u'\u276B']              # indent/dedent special chars to be used in `default` parser
    
    symbols = None      # dict of special symbols: {symbol_name: character}
    
    def __init__(self, special_chars):
        """
        :param special_chars: list of 4 unicode characters that will be used to encode INDENT_* and DEDENT_* symbols;
                              they should not occur in the script to be parsed
        """
        
        assert len(special_chars) == len(self.SPECIAL_SYMBOLS)
        self.symbols = dict(zip(self.SPECIAL_SYMBOLS, special_chars))
        
        placeholders = self.symbols.copy()
        placeholders.update({'XML_StartChar': XML_StartChar, 'XML_Char': XML_Char, 'XML_EndChar': XML_EndChar})
        
        gram = grammar % placeholders
        # print('Hypertag grammar:')
        # print(gram)
        
        super(Grammar, self).__init__(gram)
    
    @staticmethod
    def get_parser(text):
        """
        Return instance of Grammar class suitable for parsing of a given `text`.
        The grammar must be created with a proper choice of special characters,
        ones that don't collide with character set of `text`.
        """
        if not (set(Grammar.CHARS_DEFAULT) & set(text)):
            return Grammar.default
        
        chars = []
        
        # find 4 unicode characters that are not in `text`; start with CHARS_DEFAULT[0]
        code = ord(Grammar.CHARS_DEFAULT[0])
        for _ in range(4):
            while chr(code) in text:
                code += 1
            chars.append(chr(code))
            code += 1
            
        return Grammar(chars)
        
    
    def preprocess(self, text, verbose = False):
        """
        Preprocessing:
        - INDENT_* / DEDENT_* inserted in place of leading spaces/tabs
        - trailing whitespace removed in each line
        - whitespace-only lines replaced with empty lines (\n) and inserted *after* any neighboring DEDENT_*,
          so that DEDENT's always preceed empty lines and the latter can be interpreted as a top margin
          of the following block (rather than a bottom margin of the preceeding block, which would cause
          issues with proper indentation etc.)
        """
        INDENT_S = self.symbols['INDENT_S']
        DEDENT_S = self.symbols['DEDENT_S']
        INDENT_T = self.symbols['INDENT_T']
        DEDENT_T = self.symbols['DEDENT_T']

        lines = ['']            # output lines after preprocessing; empty line prepended for correct parsing of the 1st block's margin;
        linenum = 0             # current line number in input script
        current = ''            # current indentation, as a string
        margin  = 0             # current no. of empty lines that preceed the next block
        
        script = text.split('\n') + ['']            # empty line appended to ensure equal no. of DEDENT as INDENT
        total  = len(script) - 1
        
        for line in script:
            linenum += 1
            line = line.rstrip()                    # trailing whitespace removed
            tail = line.lstrip()
            indent = line[: len(line) - len(tail)]
            
            if not tail and linenum <= total:       # only whitespace in line? add to the `margin`
                margin += 1
                # lines.append('')
                
            # elif tail.startswith('--') or tail.startswith('#'):             # comment line, ignore
            #     pass
            
            else:                                   # code line, convert `indent` to INDENT_*/DEDENT_* characters and insert `tail`
                if indent == current:
                    pass

                elif indent.startswith(current):
                    increment = indent[len(current):]
                    current = indent
                    symbols = ''.join(INDENT_S if char == ' ' else INDENT_T for char in increment)
                    lines[-1] += symbols

                elif current.startswith(indent):
                    decrement = current[len(indent):]
                    current = indent
                    symbols = ''.join(DEDENT_S if char == ' ' else DEDENT_T for char in reversed(decrement))
                    lines[-1] += symbols
                    
                else:
                    raise IndentationError('indentation on line %s is incompatible with previous line' % linenum)
                    
                tail   = margin * '\n' + tail
                margin = 0
                
                lines.append(tail)
                
        assert current == '', "'%s'" % current

        # append remaining empty lines
        output = '\n'.join(lines) + margin * '\n'
        
        # drop terminal empty line that was added initially before the loop start
        assert output[-1] == '\n'
        output = output[:-1]
        
        if verbose:
            print("Grammar.preprocess() output:")
            print('-----')
            print(output)
            print('-----')

        return output
        
        
Grammar.default = Grammar(special_chars = Grammar.CHARS_DEFAULT)


########################################################################################################################################################
#####
#####  NODES
#####

class NODES(object):
    """A lexical container for definitions of all HyML tree node classes."""

    # # modes of rendering & evaluation of subtrees
    # MODE_STRICT = 1         # strict mode: requires that all variables in the subtree are defined (not MISSING), otherwise UndefinedValue exception is raised
    # MODE_NORMAL = 2         # normal mode: None variables are


    ###  BASE NODES  ###

    class node(BaseTree.node):
        indent = None               # indentation length of this node; for debugging
        
        isstatic     = False        # True in <static>, <literal> and their subclasses
        isexpression = False        # True in <expression> and subclasses - nodes that implement evaluate() method
        #iselement    = False        # True in <xelement>, <xhypertag> and other xelement subclasses
        #ishypertag   = None         # HypertagSpec object in all hypertags: <xhypertag> nodes and external hypertag objects/functions
        #isspecial    = False        # True in <special> and subclasses - nodes that mix element/hypertag functionality
        
        ispure       = None         # True if this node's render() is a pure constant function: will always return the exact same value
                                    # regardless of the context of execution and without side effects.
                                    # Is set in analyse() or compactify(), not __init__()!
        
        def check_pure(self):
            """Calculate, set and return self.ispure on the basis of check_pure() of children nodes;
            or return self.ispure if it's already set.
            """
            if self.ispure is not None: return self.ispure
            npure = sum(n.check_pure() for n in self.children)      # sum up the no. of True values among children
            self.ispure = (npure == len(self.children))             # is pure only when all children have pure=True
            return self.ispure
        
        def compactify(self, state):
            """Replace pure nodes in the subtree rooted at 'self' with static string/value nodes containg pre-computed render() result
            of a given node, so that this pre-computed string/value is returned on all future render() calls on the new node.
            Compactification is a kind of pre-rendering: whatever can be rendered in the tree before runtime variable values are known,
            is rendered and stored in the tree as static values.
            'state' is needed for render() calls because the subtree may need to push some local variables internally.
            """
            # push compactification down the tree
            for c in self.children: c.compactify(state)
            
        def compactify_self(self, state):
            "If 'self' is pure and not static, compactify it, otherwise try to compactify children. Return the new node or self."
            if self.isstatic: return self
            if self.check_pure(): return NODES.merged(self, state)
            self.compactify(state)
            return self
            
        def analyse(self, ctx):
            """
            `ctx` is an instance of Context. For read access, it can be used like a dict
            of current name->node mappings. NOTE: `ctx` contains symbols, not just raw names,
            every symbol being a concatenation of a name and a namespace qualifier (a tag or a variable),
            see TAG() and VAR() functions.
            """
            for c in self.children: c.analyse(ctx)

        # @staticmethod
        # def _translate_all(nodes, state):
        #     return DOM(n.translate(state) for n in nodes)

        @staticmethod
        def _render_all(nodes, state):
            return u''.join(n.render(state) for n in nodes)
            
        def __str__(self): return "<%s>" % self.__class__.__name__  #object.__str__(self)

    class static(node):
        """
        A node that represents static text outside expressions; its self.value is already known
        during analysis, before translate() is called. See also: class literal (for static values in expressions).
        """
        isstatic = True
        ispure   = True
        value    = None
        
        def setup(self):            self.value = self.text()
        def analyse(self, ctx):     pass
        def translate(self, state): return DOM.text(self.value) if self.value else DOM()
        def render(self, state):    return self.value
        def __str__(self):          return self.value
        
    class artificial(node):
        """Base class for artificial nodes added to the AST after parsing: during analysis or compactification."""
        type = '_artificial_'
        
        def __init__(self, origin):
            self.pos = origin.pos
            self.tree = origin.tree
            self.fulltext = origin.fulltext


    class xdocument(node):
        
        slots_in  = None    # dict of slots created for each default value to be imported into `ctx` upon startup
        slots_out = None    # dict of top-level symbols defined by this document, and their Slots: {symbol: slot}
        globals   = None    # dict of global symbols to be automatically imported into `ctx` when analysis begins
        
        # def setup(self):
        #     self.predefined = {}
        
        def predefine(self, globals):
            """
            Append `symbols` to a dict of predefined symbols that will be available to child blocks
            right from the beginning of document parsing.
            """
            if globals: self.globals = globals
        
        def analyse(self, ctx):
            self.slots_in = {symbol: ValueSlot(symbol, value, ctx) for symbol, value in self.globals.items()}
            ctx.pushall(self.slots_in)
            position = ctx.position()
            for c in self.children: c.analyse(ctx)
            self.slots_out = ctx.asdict(position)           # pull newly defined top-level symbols from the tree

        def translate(self, state):
            """
            Because <xdocument> is the root of every AST, its translate() is unusual, in that it returns a triple:
            - DOM.Root node of the final DOM generated as a result of translation of the entire AST
            - dict of top-level symbols indexed by their names: {symbol_name: value}
            - dict of top-level symbols indexed by slots: {slot: value}, for use as an initial state
              in case some hypertags get imported by other Hypertag documents and need to be expanded
              in the original (final) state of the model they were defined in
            """
            for slot in self.slots_in.values(): slot.set_value(state)
            nodes = [c.translate(state) for c in self.children]
            hroot = DOM.Root(body = nodes, indent = '\n')
            hroot.indent = ''       # fix indent to '' instead of '\n' after all child indents have been relativized
            
            # pull actual values of top-level output symbols
            symbols   = {}
            state_out = {}
            for symbol, slot in self.slots_out.items():
                try:
                    symbols[symbol] = value = slot.get(state)
                    state_out[slot] = value
                    
                except KeyError:            # some top-level slots may remain uninitialized if defined inside a control block (if/for/...)
                    continue
                    
            return hroot, symbols, state_out

        # def compactify(self, state):
        #     # if DEBUG: print("compact", "DOC", state)
        #     self.children = NODES._compactify_siblings_(self.children, state)


    ###  BLOCKS  ###

    class xblock(node):
        """Wrapper around all specific types of blocks: adds top margin and marks "outline" mode for the first returned DOM node."""
        modifier = None         # optional modifier of layout  ("<", "...") that preceeds the block
        block    = None         # the actual inner block of a specific type
        
        def setup(self):
            assert 2 <= len(self.children) <= 3 and self.children[0].type == 'margin_out'
            if len(self.children) == 3:
                self.modifier = self.children[1].value
                del self.children[1]                    # for translate() there must remain 2 children: [margin, block]
            self.block = self.children[-1]
            
        def analyse(self, ctx):
            
            # check if we're possibly moving out of the document prolog
            if ctx.in_prolog and self.block.type not in ('block_comment', 'block_context'):
                ctx.in_prolog = False

            super(NODES.xblock, self).analyse(ctx)

        def translate(self, state):
            margin, block = (c.translate(state) for c in self.children)
            if not block: return margin
            
            append = (self.modifier == '...')
            dedent = (self.modifier == '<')
            
            if not append:
                block[0].set_outline()            # mark the 1st node of the block as being "outline" not "inline"
            if dedent:
                block.set_indent('')
            return DOM(margin, block)
            
    class block_text(node):

        def translate(self, state):
            return DOM.text(self.render(state), indent = state.indentation)
            
        def render(self, state):

            # temporarily reset indentation to zero for rendering of children; this will be reverted later on
            indent = state.indentation
            state.indentation = ''
            
            lead = self.column          # horizontal position where the block's first line begins in the script
            try:
                body = self._render_all(self.children, state)
                # if body.startswith('\n'):
                #     output = body[1:]                   # no headline? drop the leading newline
                # else:
                output = ' ' * lead + body          # in the headline, spaces are prepended to replace leading tag(s) and a marker character /|!
            finally:
                state.indentation = indent

            sub_indent = get_indent(output)
            sub_indent = sub_indent[:lead+1]        # max 1 initial space/tab after the lead is dropped; remaining sub-indentation is preserved in `output`
            output = del_indent(output, sub_indent)
            # print('sub_indent: "{sub_indent}"')

            # if tail lines have shorter indent than the headline, drop all the lead + 1 space (gap)
            if len(sub_indent) < lead:
                drop = lead - len(sub_indent)
                if output[drop:drop+1] == ' ': drop += 1
                output = output[drop:]
                
            # # drop the 1st line (headline) if empty
            # if output.split('\n', 1)[0].strip() == '':
            #     output =
            
            return output

    class xblock_markup(block_text): pass
    class xblock_normal(block_text): pass
    class xblock_verbat(block_text): pass
    class xblock_comment(block_text):
        def translate(self, state): return DOM()
        def render(self, state):    return ""
    
    class xblock_embed(node):
        """Embedding of DOM nodes through @... type of expression."""
        expr = None
        
        def setup(self):
            self.expr = self.children[0]
            
        def translate(self, state):
            body = self.expr.evaluate(state)
            body = self._as_sequence(body)
            body.set_indent(state.indentation)                  # set indentation of the fragment to be inserted
            return body
        
        def _as_sequence(self, body):
            if isinstance(body, DOM): return body
            if isinstance(body, DOM.Node): return DOM(body)
            try:
                body = list(body)
            except Exception as ex:
                raise TypeErrorEx("embedded @-expression evaluates to %s instead of a DOM or DOM.Node or an iterable of DOM.Node's" % type(body), self)
            return DOM(*body)

    class xblock_struct(node):
        tags = None         # <tags_expand> node
        body = None         # <body_struct> node
        
        def setup(self):
            self.tags = self.children[0]
            self.body = self.children[1]
            assert self.tags.type == 'tags_expand'
            
        def analyse(self, ctx):
            ctx.regular_depth += 1
            position = ctx.position()
            for c in self.children: c.analyse(ctx)
            ctx.reset(position)             # tagged node defines a local namespace, hence need to drop symbols defined inside
            ctx.regular_depth -= 1
            
        def translate(self, state):
            body = self.body.translate(state)
            body = self.tags.apply_tags(state, body)
            body.set_indent(state.indentation)
            return body

    class xblock_def(node, Hypertag):
        """Definition of a native hypertag."""
        name       = None
        attrs      = None           # all attributes as a list of children nodes, including @body
        attr_body  = None           # the @body node if present, otherwise None
        attr_names = None           # names of all attributes, including @body, as a dict {name: node}
        attr_regul = None           # all regular (non-body) attributes, as a list of children
        body       = None
        slot       = None
        native     = None           # a Native tag instance that will be inserted into all DOMs
        
        def setup(self):
            self.name  = self.children[0].value
            self.attrs = self.children[1:-1]
            self.body  = self.children[-1]
            assert all(attr.type.startswith('attr_') for attr in self.attrs)
            
            self.attr_names = {attr.name: attr for attr in self.attrs}

            # TODO: check that attr names are Python identifiers (name_id), otherwise they can't be used in expressions
            
            # check that attr names are unique
            if len(self.attr_names) < len(self.attrs):
                raise SyntaxErrorEx("duplicate attribute '%s' in hypertag definition '%s'" % (duplicate(self.attrs), self.name), self)
            
            # pick body attribute (attr_body) and the remaining list of regular attributes (attr_regul)
            if self.attrs and self.attrs[0].body:
                self.attr_body  = self.attrs[0]
                self.attr_regul = self.attrs[1:]
            else:
                self.attr_regul = self.attrs
            
            # create a Native tag that will be inserted into DOMs
            self.native = Native(self.name)
            
        def analyse(self, ctx):
            if ctx.control_depth >= 1: raise SyntaxErrorEx('hypertag definition inside a control block is not allowed', self)
            
            ctx.regular_depth  += 1
            ctx.hypertag_depth += 1
            position = ctx.position()
            
            for attr in self.attrs:                 # analyse default-value expressions of attributes
                attr.analyse(ctx)

            for attr in self.attrs:                 # declare attributes as local variables for subsequent analysis of self.body
                attr.declare_var(ctx)

            self.body.analyse(ctx)                  # analyse the formal body

            ctx.reset(position)
            ctx.hypertag_depth -= 1
            ctx.regular_depth  -= 1

            symbol = TAG(self.name)
            self.slot = ValueSlot(symbol, self, ctx)
            ctx.push(symbol, self.slot)
            
        def translate(self, state):
            self.slot.set_value(state)
            return None                 # hypertag produces NO output in the place of its definition (only in places of occurrence)

        def expand(self, body, attrs, kwattrs, state, caller):
            """
            Translate the formal self.body in a given `state`, insert the actual `body` wherever necessary,
            and return as a DOM (not a string!) for possible further manipulation in other hypertags.
            """
            dom_attrs = self._append_attrs(body, attrs, kwattrs, state, caller)         # extend `state` with actual values of tag attributes
            output = self.body.translate(state)
            output.set_indent(state.indentation)
            # return output
            
            if output: output[0].set_outline(False)         # node's `outline` will be set for the root node up in xblock.translate()
            return DOM.node(output, state.indentation, tag = self.native, kwattrs = dom_attrs)

        def _append_attrs(self, body, attrs, kwattrs, state, caller):
            """Extend `state` with actual values of tag attributes."""

            # verify no. of positional attributes & names of keyword attributes
            if len(attrs) > len(self.attr_regul):
                raise TypeErrorEx("hypertag '%s' takes %s positional attributes but %s were given" % (self.name, len(self.attr_regul), len(attrs)), caller)
            if self.attr_body and self.attr_body.name in kwattrs:
                raise TypeErrorEx("direct assignment to body attribute '%s' of hypertag '%s' is not allowed" % (self.attr_body.name, self.name), caller)
            
            # translate attribute names in `kwattrs` to nodes as keys
            try:
                kwattrs = {self.attr_names[name]: value for name, value in kwattrs.items()}
            except KeyError as ex:
                name = ex.args[0]
                raise TypeErrorEx("hypertag '%s' got an unexpected keyword attribute '%s'" % (self.name, name), caller)

            dom_attrs = {}
            dom_kwattrs = {attr.name: value for attr, value in kwattrs.items()}
            
            # move positional attributes to `kwattrs`
            for pos, value in enumerate(attrs):
                attr = self.attr_regul[pos]
                if attr in kwattrs: raise TypeErrorEx("hypertag '%s' got multiple values for attribute '%s'" % (self.name, attr.name), caller)
                kwattrs[attr] = value
                dom_attrs[attr.name] = value

            dom_attrs.update(dom_kwattrs)
            # dom_attrs = {attr.name: value for attr, value in kwattrs.items()}
            
            # impute missing values with defaults
            for attr in self.attr_regul:
                if attr not in kwattrs:
                    if attr.expr is None: raise TypeErrorEx("hypertag '%s' missing a required positional attribute '%s'" % (self.name, attr.name), caller)
                    kwattrs[attr] = attr.expr.evaluate(state)
                    
            # transfer attribute values from `kwattrs` to `state`
            for attr, value in kwattrs.items():
                attr.assign(state, value)
            # state.update(kwattrs)
            
            # append `body` to `state`
            if self.attr_body:
                self.attr_body.assign(state, body)
                # state[self.attr_body] = body
            elif body:
                raise VoidTagEx("non-empty body passed to a void hypertag '%s'" % self.name, caller)
            
            return dom_attrs
            
    class xblock_import(node):
        """"""
        path  = None            # import path string; optional
        items = None            # list of 1+ nodes of type <xwild_import> or <xname_import>
        
        def setup(self):
            if self.children[0].type == 'path_import':
                self.path  = self.children[0].value
                self.items = self.children[1:]
            else:
                self.items = self.children
            for item in self.items:
                item.path = self.path
                
        def analyse(self, ctx):
            if ctx.control_depth >= 1: raise SyntaxErrorEx('import inside a control block is not allowed', self)
            super(NODES.xblock_import, self).analyse(ctx)

        def translate(self, state):
            for item in self.items: item.translate(state)
            return DOM()
        
    class _import(node):
        def _import(self, path):
            runtime = self.tree.runtime
            module  = self.tree.module
            return runtime.import_module(path, module, self)
        
    class xwild_import(_import):
        path    = None          # path string as specified in the "from" clause
        slots   = None          # dict of symbols and their slots created during analysis
        
        def analyse(self, ctx):
            module = self._import(self.path)

            # exclude private symbols AND wrap up all native tag definitions within ImportedHypertag
            # to preserve original runtime state of their module, for expand()
            symbols = {name: ImportedHypertag(value, module.state) if (IS_TAG(name) and isinstance(value, Hypertag)) else value
                       for name, value in module.symbols.items() if name[1] != '_'}
            
            self.slots = {symbol: ValueSlot(symbol, value, ctx) for symbol, value in symbols.items()}
            ctx.pushall(self.slots)

        def translate(self, state):
            for slot in self.slots.values(): slot.set_value(state)
            
    class xname_import(_import):
        path  = None        # path string as specified in the "from" clause
        slot  = None        # <slot> that will keep value of this imported symbol
        
        def analyse(self, ctx):
            assert self.path
            symbol  = self.children[0].value                        # original symbol name with leading % or $
            rename  = (symbol[0] + self.children[1].value) if len(self.children) == 2 else symbol
            module  = self._import(self.path)

            if symbol not in module.symbols: raise ImportErrorEx("cannot import '%s' from a given path (%s)" % (symbol, self.path), self)
            value = module.symbols[symbol]
            
            if IS_TAG(symbol) and isinstance(value, Hypertag):
                value = ImportedHypertag(value, module.state)
            
            self.slot = ValueSlot(rename, value, ctx)
            ctx.push(rename, self.slot)

        def translate(self, state):
            self.slot.set_value(state)

    class xblock_context(xblock_import):
        """
        Wrapper-block around `context` imports (xcntx_import). Checks whether the block is placed
        at the beginning of a document (in document's prolog).
        """
        def analyse(self, ctx):
            if not ctx.in_prolog: raise SyntaxErrorEx('context block(s) must be the first block(s) in a document (after comments)', self)
            super(NODES.xblock_import, self).analyse(ctx)

    class xcntx_import(node):
        """Like <xname_import> but imports the name from a dynamic context rather than from a module."""
        slot  = None        # <slot> that will keep value of the symbol
        
        def analyse(self, ctx):
            context = self.tree.context
            symbol  = self.children[0].value                            # original symbol name with leading % or $
            rename  = (symbol[0] + self.children[1].value) if len(self.children) == 2 else symbol
            
            if symbol not in context: raise ImportErrorEx("symbol '%s' not found in context" % symbol, self)
            value = context[symbol]
            self.slot = ValueSlot(rename, value, ctx)
            ctx.push(rename, self.slot)

        def translate(self, state):
            self.slot.set_value(state)

        
    class control_block(node):
        """
        Base class for if/try/for/while blocks. Inside these blocks, the execution may proceed through 2 or more different
        paths (one of them being the "null" branch) and for this reason they require special handling during analysis.
        """
        # slot_maps = None        # list of slot mappings for every branch including the null branch (-1); each mapping
        #                         # is a dict of {input_or_local_slot: output_slot} pairs; values are copied
        #                         # from input/local to output slots after translation of the block
        
        def analyse(self, ctx):
            ctx.control_depth += 1
            self._analyse_branches(ctx)
            ctx.control_depth -= 1

        def _analyse_branches(self, ctx):
            """
            Unlike a regular tagged block, if/try block does NOT introduce a new namespace,
            so all symbols defined in branches must be made visible to sibling nodes that go after the block.
            Moreover, if a variable is (re)declared multiple times in separate branches, all these declaraions
            (i.e., their corresponding slots) must be mapped onto a single shared outgoing slot.
            Also, if any of the locally-defined variables overrides an existing variable defined before the block,
            the incoming variable must be mapped onto an outgoing slot, as well, for proper handling of a "null branch" execution
            (i.e., the situation when no code of this block gets executed, like in an empty loop or a failed "try" block).
            """
            # # list of dicts of symbols and local slots defined in each subsequent branch, including the "null" branch at [-1]
            # local_slots = []
            # new_symbols = []
            
            for branch in self.children:
                position = ctx.position()
                branch.analyse(ctx)
                symbols = ctx.asdict(position)          # top-level symbols declared in this branch...
                ctx.reset(position)
                ctx.pushnew(symbols)                    # ...only new symbols (not declared in a previous branch) are added
                
                # local_slots.append(symbols)
                # new_symbols += symbols.keys()
                
            # if not new_symbols: return
            # new_symbols = set(new_symbols)
            #
            # # [-1] contains symbols of the "null" branch, they will be copied from incoming slots when no clause is executed
            # null_slots = {symbol: ctx[symbol] for symbol in new_symbols if symbol in ctx}
            # local_slots.append(null_slots)
            #
            # # create outgoing slots for all newly defined variables
            # output_slots = {symbol: Slot(symbol, ctx) for symbol in new_symbols}
            #
            # mapping_default = {input_slot: output_slots[symbol] for symbol, input_slot in null_slots.items()}
            #
            # def match(slots):
            #     mapping_branch = {local_slot: output_slots[symbol] for symbol, local_slot in slots.items()}
            #     mapping = mapping_default.copy()
            #     mapping.update(mapping_branch)
            #     return mapping
            #
            # # set up mappings from local/incoming slots to outgoing slots for each branch separately including the "null" branch
            # self.slot_maps = list(map(match, local_slots))
            #
            # ctx.pushall(output_slots)
            
        # def _fill_slots(self, branch, state):
        #     """Fill out outgoing slots according to what execution branch (0..N-1) has been selected; -1 is the "null" branch.
        #     """
        #     # for slot in self.output_slots:
        #     #     slot.set_value(branch, state)
        #     mapping = self.slot_maps[branch]
        #     for in_slot, out_slot in mapping.items():
        #         value = in_slot.get(state)
        #         out_slot.set(state, value)
        
        def _select_branch(self, state):
            raise NotImplementedError
        
        def translate(self, state):
            body, branch = self._select_branch(state)
            # self._fill_slots(branch, state)
            body.set_indent(state.indentation)
            return body
        

    class xblock_try(control_block):
        """
        A "try" block. Two syntax forms available:
        - short form:   ?tag... ?|...
        - long form:    try ... else ... else ...
        Every "try" block catches ALL exceptions that inherit from Exception class, including all Hypertag exceptions.
        There is no way to explicitly restrict the scope of exceptions caught (no "except" clause), unlike in Python.
        The block does NOT catch special exceptions that inherit directly from BaseException:
        SystemExit, KeyboardInterrupt, GeneratorExit.
        Note that the meaning of the "else" clause is OPPOSITE to what it is in Python: here, "else" branch
        is executed if all preceeding try/else branches failed with exceptions.
        If all branches fail, including the last one, empty output is rendered (exception from the last branch is NOT re-raised).
        """
        def _select_branch(self, state):
            for i, branch in enumerate(self.children):
                try:
                    return branch.translate(state), i
                except Exception as ex:
                    pass
            return DOM(), -1
    
    class xblock_if(control_block):
        clauses  = None         # list of 1+ <clause_if> nodes
        elsebody = None         # optional <body_*> node for the "else" branch
        
        def setup(self):
            if self.children[-1].type == 'clause_if':
                self.clauses = self.children
            else:
                self.clauses = self.children[:-1]
                self.elsebody = self.children[-1]

        def _select_branch(self, state):
            for i, clause in enumerate(self.clauses):
                if clause.test.evaluate(state):
                    return clause.translate(state), i
            if self.elsebody:
                return self.elsebody.translate(state), -2
            return DOM(), -1
        
    class xclause_if(node):
        test = None             # <expression> node containing a test to be performed
        body = None             # <body> to be rendered if the clause is positive
        def setup(self):
            assert 1 <= len(self.children) <= 2
            self.test = self.children[0]
            self.body = self.children[1] if len(self.children) == 2 else None
        def translate(self, state):
            return self.body.translate(state) if self.body else DOM()
        # def render(self, state):
        #     return self.body.render(state)
        
    class xclause_else(node):
        def translate(self, state):
            assert len(self.children) == 1
            return self.children[0].translate(state)
        
    class xblock_while(control_block):
        # def _analyse_branches(self, ctx):
        #     self.children[0].analyse(ctx)
        def translate(self, state):
            out = []
            empty = True
            clause = self.children[0]
            else_  = self.children[1] if len(self.children) > 1 else None
            
            while clause.test.evaluate(state):
                empty = False
                body = clause.translate(state)
                out += body.nodes

            # use the `else` clause if `sequence` was empty
            if empty and else_:
                out = else_.translate(state)
            else:
                out = DOM(*out)

            out.set_indent(state.indentation)
            return out
        
    class xblock_for(control_block):
        targets = None              # 1+ loop variables to assign to
        expr    = None              # loop expression that returns a sequence (iterable) to be looped over
        body    = None              # inner blocks of the loop, can be missing
        else_   = None              # "else" clause, optional
        
        def setup(self):
            self.targets, self.expr = self.children[:2]
            self.body = self.children[2] if len(self.children) >= 3 else None
            self.else_ = self.children[-1] if self.children[-1].type == 'clause_else' else None
            assert isinstance(self.expr, NODES.expression)
            assert self.targets.type == 'targets'
            # assert self.targets.type == 'var', 'Support for multiple targets in <for> not yet implemented'
            
        def _analyse_branches(self, ctx):
            self.expr.analyse(ctx)
            self.targets.analyse(ctx)
            if self.body: self.body.analyse(ctx)

        def translate(self, state):
            out = []
            empty = True
            sequence = self.expr.evaluate(state)
            
            for value in sequence:                  # translate self.body multiple times, once for each value in `sequence`
                empty = False
                self.targets.assign(state, value)
                if self.body:
                    body = self.body.translate(state)
                    out += body.nodes
                
            # use the `else` clause if `sequence` was empty
            if empty and self.else_:
                out = self.else_.translate(state)
            else:
                out = DOM(*out)
                
            out.set_indent(state.indentation)
            return out

    class xblock_assign(node):
        targets = None
        inplace = None          # symbol of in-place arithmetic operator to apply (+-*/), optional
        expr    = None
        oper    = None          # 2-arg function that implements `inplace` operator
        
        opers   = {'+':  operator.add,
                   '-':  operator.sub,
                   '*':  operator.mul,
                   '%':  operator.mod,
                   '/':  getattr(operator, 'div', None) or operator.truediv,
                   '//': operator.floordiv,
                   '<<': operator.lshift,
                   '>>': operator.rshift,
                   '&':  operator.and_,
                   '|':  operator.or_,
                   '^':  operator.xor,
        }
        
        def setup(self):
            assert 2 <= len(self.children) <= 3
            self.targets  = self.children[0]
            self.expr     = self.children[-1]

            # in-place operator-assignment ?
            if len(self.children) == 3:
                inplace   = self.children[1].value
                self.oper = self.opers[inplace]
                if self.targets.is_augmented():
                    raise SyntaxErrorEx("illegal expression %s= for augmented assignment" % inplace, self)
                self.targets.set_inplace()
            
        def analyse(self, ctx):
            self.expr.analyse(ctx)
            self.targets.analyse(ctx)

        def translate(self, state):
            value = self.expr.evaluate(state)
            if self.oper:
                current = self.targets.evaluate(state)
                value   = self.oper(current, value)
            self.targets.assign(state, value)
            return None
    
    class xtargets(node):
        def is_augmented(self):
            """True if self represents multiple targets (not a plain variable) to assign to."""
            return len(self.children) >= 2 or self.children[0].type != 'var_def'
        
        def set_inplace(self):
            self.children[0].set_inplace()
            
        def analyse(self, ctx):
            """Recursively insert all variables that comprise this target into context."""
            for c in self.children: c.analyse(ctx)

        def assign(self, state, value):
            """Unpack `value` and assign to child targets."""
            N = len(self.children)
            if N == 1:
                self.children[0].assign(state, value)
                return
            
            # unpack and assign to multiple child targets
            i = 0
            for i, v in enumerate(value):               # raises TypeError if `value` is not iterable
                if i >= N: raise ValueErrorEx("too many values to unpack (expected %s)" % N, self)
                self.children[i].assign(state, v)
            if i+1 < N:
                raise ValueErrorEx("not enough values to unpack (expected %s, got %s)" % (N, i+1), self)
            
        def evaluate(self, state):
            assert len(self.children) == 1
            return self.children[0].evaluate(state)
            
            
    ###  BODY & LINES  ###

    class body(node):
        def translate(self, state):
            return DOM(n.translate(state) for n in self.children)
            # return self._translate_all(self.children, state)
    
    class xbody_control(body): pass
    class xbody_struct (body): pass

    class line(node):
        def translate(self, state):
            return DOM.text(self.render_inline(state))
        def render(self, state):
            return state.indentation + self.render_inline(state)
        def render_inline(self, state):
            """Render contents of the line, i.e., everything except boundary (indentation, margin). Implemented by subclasses."""
            raise NotImplementedError

    class xline_verbat(line):
        def render_inline(self, _):
            return self.text()

    class xline_normal(line):
        def render_inline(self, state):
            assert len(self.children) == 1
            child = self.children[0]
            assert child.type == 'line_markup'
            text = child.render_inline(state)                   # this calls xline_markup.render_inline()
            # escape = self.tree.config['escape_function']
            escape = self.tree.runtime.escape
            return escape(text) if escape else text

    class xline_markup(line):
        def render_inline(self, state):
            markup = self._render_all(self.children, state)     # renders embedded expressions, in addition to static text
            return markup

    
    ###  TAGS & HYPERTAGS  ###

    class xtags_expand(node):
        """List of tag_expand nodes."""
        def apply_tags(self, state, body):
            """Wrap up the translated `body` (a DOM, not AST) with subsequent tags processed (expanded) in reverse order."""
            for tag in reversed(self.children):
                body = tag.translate_tag(state, body)
            return body
        
    class xtag_expand(node):
        """
        Occurrence of a tag.
        NOTE #1: unnamed attrs can be *mixed* with named ones (unlike in python) -
                 during tag expansion all unnamed attrs are passed first to a tag, followed by all named (keyword-) ones
        NOTE #2: same attr can appear more than once, in such case its values (must be strings!) get space-concatenated;
                 this is particularly useful for "class" attibute and its short form:  div .top.left.darkbg
        """
        DEFAULT = "div"     # default `name` when no tag name was provided (a shortcut was used: .xyz or #xyz); UNUSED!
        name  = None        # tag name: a, A, h1, div ...
        tag   = None        # resolved definition of this tag, as a Tag or Hypertag (xblock_def) instance; or a Slot
        attrs = None        # 0+ list of <attr_short> and <attr_val> nodes
        unnamed = None      # list of <expression> nodes of unnamed attributes from `attrs`
        named   = None      # list of (name, expression) pairs of named attributes from `attrs`; duplicate names allowed
        
        def setup(self):
            
            # retrieve `name` of this tag
            head = self.children[0]
            if head.type == 'name_id':
                self.name = head.value
                self.attrs = self.children[1:]
            else:
                self.name = self.DEFAULT
                self.attrs = self.children
                
            self.unnamed = []
            self.named = []
            
            # collect attributes: their names (optional) and expressions (obligatory);
            for attr in self.attrs:
                name = attr.name
                expr = attr.expr
                if name is None:
                    if self.named: raise SyntaxErrorEx("positional argument follows keyword argument")
                    self.unnamed.append(expr)
                else:
                    # if name in self.named: raise SyntaxErrorEx("attribute '{name}' appears twice on attributes list of tag '{self.name}'", attr)
                    self.named.append((name, expr))
                
        def analyse(self, ctx):
            
            for c in self.attrs: c.analyse(ctx)
            self.tag = ctx.get(TAG(self.name))
            if self.tag is None: raise UndefinedTagEx("undefined tag '%s'" % self.name, self)
            
        def translate_tag(self, state, body):
            """
            translate_tag() differs from a regular translate() in that it accepts `body` additionaly.
            The actual `body` is already translated and has a form of a DOM.
            """
            attrs, kwattrs = self._eval_attrs(state)            # calculate actual values of attributes
            
            assert isinstance(self.tag, Slot)
            tag = self.tag.get(state)

            if isinstance(tag, Hypertag):
                return tag.expand(body, attrs, kwattrs, state, self)
            
            elif isinstance(tag, Tag):
                return DOM.node(body, tag = tag, attrs = attrs, kwattrs = kwattrs)
            
            else:
                raise NotATagEx("Not a tag: '%s' (%s)" % (self.name, tag.__class__), self)
            
        def _eval_attrs(self, state):
            unnamed = [attr.evaluate(state) for attr in self.unnamed]
            
            named = {}
            for name, expr in self.named:
                value = expr.evaluate(state)
                if name in named:
                    named[name] += ' ' + value       # = '{named[name]} {value}'
                else:
                    named[name] = value
                    
            return unnamed, named
            
    class xnull(node):
        def translate_tag(self, state, body):
            return DOM.node(body, tag = null)
        
    class xpass(node):
        def translate(self, state):
            return DOM()
        

    ###  ATTRIBUTES & ARGUMENTS  ###
    
    class attribute(node):
        """Attribute inside a hypertag occurrence OR tag definition:
            unnamed / named / short (only in tag occurence) / obligatory / body (only in tag definition).
        """
        name = None         # [str] name of this attribute; None if unnamed
        expr = None         # <expression> node of this attribute; None if no expression present (attr definition with no default)
        body = False        # True in xattr_body
        slot = None         # Slot that identifies this variable inside `state`
        
        def declare_var(self, ctx):
            symbol = VAR(self.name)
            self.slot = Slot(symbol, ctx)
            ctx.push(symbol, self.slot)
        
        def assign(self, state, value):
            self.slot.set(state, value)
        
    # in-definition attributes:  xattr_body, xattr_def
    
    class xattr_body(attribute):
        body = True
        def setup(self):
            assert len(self.children) == 1
            self.name = self.children[0].value          # <name_id>
            
    class xattr_def(attribute):
        def setup(self):
            assert 1 <= len(self.children) <= 2
            self.name = self.children[0].value          # <name_xml>
            if len(self.children) == 2:
                self.expr = self.children[1]
            
    # in-occurrence attributes:  xattr_named, xattr_unnamed, xattr_short
    
    class xattr_named(attribute):
        def setup(self):
            assert len(self.children) == 2
            self.name = self.children[0].value          # <name_xml>
            self.expr = self.children[1]
            
    class xattr_unnamed(attribute):
        def setup(self):
            assert len(self.children) == 1
            self.expr = self.children[0]
            
    class xattr_short(attribute):
        def setup(self):
            symbol = self.fulltext[self.pos[0]]
            assert symbol in '.#'
            self.name = 'class' if symbol == '.' else 'id'
            assert len(self.children) == 1
            self.expr = self.children[-1]
        
    class xkwarg(node):
        name = None
        expr = None
        
        def setup(self):
            assert len(self.children) == 2
            assert self.children[0].type == 'name_id'
            self.name = self.children[0].value
            self.expr = self.children[1]
            
        def evaluate(self, state):
            return self.expr.evaluate(state)
        

    ###  EXPRESSIONS - ROOT NODES  ###
    
    class expression(node):
        """Base class for all nodes that represent an expression, or its part (a subexpression)."""
        
        qualifier = None            # optional qualifier: ? or ! ... used only in a few node types
        
        def evaluate(self, state):
            raise NotImplementedError

        def evaluate_with_qualifier(self, state):
            """Special variant of evaluate() to be used in these expression nodes that may have a not-None qualifier.
               They should call this method in evaluate() and implement _eval_inner_qualified().
            """
            try:
                val = self._eval_inner_qualified(state)
            except Exception as ex:
                if self.qualifier == '?': return ''
                else: raise
            
            if val: return val
            
            # `val` is false ... check qualifiers to undertake appropriate action
            if self.qualifier == '?': return ''
            if self.qualifier == '!': raise FalseValueEx("Obligatory expression evaluates to a false value (%s)" % repr(val), self)
            return val

        def _eval_inner_qualified(self, state):
            raise NotImplementedError


    class expression_root(expression):
        """Base class for root nodes of all non-literal embedded expressions, either in markup or attribute/argument lists.
        """
        qualifier = None        # optional qualifier: ? or !
        context   = None        # copy of Context that has been passed to this node during analyse(); kept for re-use by render(),
                                # in case if the expression evaluates to yet another (dynamic) piece of Hypertag code
        def setup(self):
            # see if there is a qualifier added as a sibling of this node
            if self.sibling_next and self.sibling_next.type == 'qualifier':
                self.qualifier = self.sibling_next.text()
        
        def render(self, state):
            """Rendering is invoked only for a root node of an expression embedded in xline_* node of text."""
            return STR(self.evaluate(state), self)
        
        def evaluate(self, state):
            return self.evaluate_with_qualifier(state)
        
        def _eval_inner_qualified(self, state):
            assert len(self.children) == 1
            return self.children[0].evaluate(state)

    class xexpr(expression_root): pass
    class xexpr_var(expression_root): pass
    class xexpr_factor(expression_root): pass
    class xexpr_strict(expression_root): pass
    class xexpr_augment(expression_root): pass
    class xexpr_bitwise(expression_root): pass
    
    class xtag_use(expression):
        """Occurrence of a tag inside an expression: %TAG."""
        symbol = None
        slot   = None           # Slot that identifies this tag inside `state` for read access
        
        def setup(self):
            self.symbol = self.text()
        
        def analyse(self, ctx):
            self.slot = ctx.get(self.symbol)
            if self.slot is None: raise NameErrorEx("tag '%s' is not defined" % self.symbol, self)
            assert isinstance(self.slot, Slot), "not a slot: %s in %s" % (type(self.slot), self.symbol)
            
        def evaluate(self, state):
            try:
                tag = self.slot.get(state)
                if isinstance(tag, Hypertag):
                    return EmbeddedHypertag(tag, state, self)
                else:
                    return tag
                
            except KeyError:
                raise UnboundLocalEx("tag '%s' referenced before declaration or assignment" % self.symbol, self)


    class variable(expression):
        """Common code for both a variable's definition (xvar_def) and a variable's occurrence (xvar_use)."""
        name  = None
        read  = False       # True in <xvar_use> and this <xvar_def> which performs in-place operation
        write = False       # True in <xvar_def> and <attribute>
        
        slot_read  = None       # Slot that identifies this variable inside `state` for read access
        slot_write = None       # Slot that identifies this variable inside `state` for write access
        
        # slot_read & slot_write may reference different Slots, e.g., in a reassignment block that overrides a name defined upper in the doc:
        #   $x = 1
        #   div:
        #      $x += 1        -- this <variable> reads "x" from the upper level and assigns a new "x" in a narrower name scope, hence a new Slot is needed
        
        def setup(self):
            self.name = self.text()
            if self.name[0] == MARK_VAR:            # preceeding $ is allowed inside {...} and should be truncated here
                self.name = self.name[1:]
        
        def analyse(self, ctx):
            symbol = VAR(self.name)
            slot   = ctx.get(symbol)
            assert slot is None or isinstance(slot, Slot), "not a slot: %s in %s" % (type(slot), self.name)

            if self.read:
                if slot is None: raise NameErrorEx("variable '%s' is not defined" % self.name, self)
                self.slot_read = slot
            
            if self.write:
                if slot and slot.depth == ctx.regular_depth:
                    self.slot_write = slot
                else:
                    self.slot_write = Slot(symbol, ctx)
                    ctx.push(symbol, self.slot_write)
            
        def evaluate(self, state):
            
            assert self.read
            try:
                return self.slot_read.get(state)
            
            except KeyError:
                raise UnboundLocalEx("variable '%s' referenced before assignment" % self.name, self)

        def assign(self, state, value):
            assert self.write
            self.slot_write.set(state, value)

    class xvar_use(variable):
        """Occurence (use) of a variable."""
        read = True

    class xvar_def(variable):
        """Definition of a variable, or assignment to a previously defined variable."""
        write = True
        
        def set_inplace(self):
            """Mark this definition as occuring in an in-place assignment (+= -= etc.) so the variable will be
               used in read access, not only for writing.
            """
            self.read = True
        

    ###  EXPRESSIONS - TAIL OPERATORS  ###
    
    class tail(node):
        """Tail operators implement apply() instead of evaluate()."""
        def apply(self, obj, state):
            raise NotImplementedError

    class xcall(tail):
        title = 'function call (...)'                   # for error messaging
        def apply(self, fun, state):
            args, kwargs = self._evaluate_args(state)
            return fun(*args, **kwargs)
        
        def _evaluate_args(self, state):
            items  = [(c.name if c.type == 'kwarg' else None, c.evaluate(state)) for c in self.children]
            args   = [value       for name, value in items if name is None]
            kwargs = {name: value for name, value in items if name is not None}
            return args, kwargs
            
    class xpartial_call(xcall):
        """Partial function call inside pipeline operator, like in x:fun(a,b)."""
        def apply(self, fun, state):
            if not self.children: return fun
            args, kwargs = self._evaluate_args(state)
            return partial(fun, *args, **kwargs)
            
    class xslice_value(expression):
        def evaluate(self, state):
            assert len(self.children) <= 1
            if self.children: return self.children[0].evaluate(state)
            return None                                 # None indicates an empty index, like in 1:, in the slice(...) object
            
    class xindex(tail):
        """Element access: [...], with any type of subscript: [i], [i:j], [i:j:k], [::] etc.
        Children after reduction are either a single <xexpr> node (no slicing),
        or a list of 2-3 <xslice_value> nodes in case of a slice.
        """
        title = 'sequence index [...]'

        def compactify(self, state):
            assert 1 <= len(self.children) <= 3
            self.children = [n.compactify(state) for n in self.children]
            return self

        def apply(self, obj, state):
            # simple index: [i]
            if len(self.children) == 1:
                index = self.children[0].evaluate(state)
                return obj[index]
            
            # 2- or 3-element slice index:  i:j[:k]
            values = [n.evaluate(state) for n in self.children]
            return obj[slice(*values)]
        
    class xmember(tail):
        title = 'member access "."'
        def compactify(self, state):
            return self                                 # no compactification, it's only 1 child: a static identifier
        def apply(self, obj, state):
            assert self.children[0].type == "name_id"
            member = self.children[0].value
            return getattr(obj, member)
    
    class xqualifier(static):
        def setup(self): self.value = ''

    class xfactor(expression):
        """A chain of tail operators: () [] . with optional trailing qualifier ? or ! """
        atom      = None
        tail      = None        # list of tail operators: call / index / member
        qualifier = None        # optional qualifier: ? or !
        
        def setup(self):
            self.atom = self.children[0]
            self.tail = self.children[1:]
            if self.tail and self.tail[-1].type == 'qualifier':
                qualifier_node = self.tail.pop()
                self.qualifier = qualifier_node.text()
            
        def evaluate(self, state):
            return self.evaluate_with_qualifier(state)
        
        def _eval_inner_qualified(self, state):
            val = self.atom.evaluate(state)
            for op in self.tail:
                assert isinstance(op, NODES.tail)
                val = op.apply(val, state)
            return val
    
    class xfactor_var(xfactor): pass
    class xfactor_filt(xfactor): pass
    class xfactor_strict(xfactor): pass
    class xfilter(xfactor): pass
    
    
    ###  EXPRESSIONS - OPERATORS (BINARY / TERNARY)  ###

    class static_operator(static):
        name  = None        # textual representation of the operator, for possible rendering back into the document
        apply = None        # corresponding function from 'operator' module
        
        ops = ['+ add', '- sub', '** pow', '* mul', '// floordiv', '% mod', '<< lshift', '>> rshift', '& and_', '| or_', '^ xor',
               '< lt', '> gt', '== eq', '>= ge', '<= le', '!= ne', 'is is_', 'is not is_not']
        ops = [m.rsplit(' ', 1) for m in ops]
        ops = {op: getattr(operator, fun) for op, fun in ops}
        
        # '/' must be added separately, because it has different names (and behavior) in Python 2 vs. 3
        ops['/'] = getattr(operator, 'div', None) or operator.truediv
        
        # extra operators, implemented by ourselves
        ops['in'] = lambda x, d: x in d                         # operator.contains() is not suitable bcs it takes operands in reversed order
        ops['not in'] = lambda x, d: x not in d
        ops[''] = ops['+']                                      # missing operator mapped to '+' (implicit +)
        
        def setup(self):
            self.name = self.text()
            self.name = ' '.join(self.name.split())             # to replace multiple whitespaces in "not in", "is not"
            self.apply = self.ops[self.name]
            
    class xop_multiplic(static_operator): pass
    class xop_additive(static_operator): pass
    class xop_power(static_operator): pass
    class xop_shift(static_operator): pass
    class xop_comp(static_operator): pass
    class xop_inplace(static): pass
    class xneg(static): pass                            # negation is implemented inside <xarith_expr>
    class xnot(static): pass                            # static keyword "not" must have a node, for counting of repeated "not not not ..." expression
    
    class chain_expression(expression):
        """A chain of different binary operators, all having the same priority: x1 OP1 x2 OP2 x3 ..."""

        def evaluate(self, state):
            head, tail = self._prepare(state)
            ops = tail[0::2]                            # items 0,2,4,... are operators
            exprs = tail[1::2]                          # items 1,3,5,... are subsequent expressions, after the initial one
            assert len(exprs) == len(ops)
            
            res = head
            for op, expr in zip(ops, exprs):                # adding terms one by one to 'res'
                val = expr.evaluate(state)
                res = op.apply(res, val)                    # calulate: <res> = <res> op <val>
            
            return res
        
        def _prepare(self, state):
            """Pre-processesing of the 1st item of the chain for evaluate(). Returns the chain as (head, tail) for actual evaluation.
            Override in subclasses if the 1st item is treated differently then the others."""
            head = self.children[0].evaluate(state)
            tail = self.children[1:]
            return head, tail
    
    class xpow_expr(chain_expression):
        """chain of power operators: ** """
    class xterm(chain_expression):
        """chain of multiplicative operators: * / // %"""
    class xshift_expr(chain_expression):
        """chain of shift operators: << >>"""
    class xarith_expr(chain_expression):
        """chain of additive operators: neg + -"""
        def _prepare(self, state):
            if self.children[0].type == 'neg':
                head = self.children[1].evaluate(state)
                if head is not None:
                    head = -head
                tail = self.children[2:]
            else:
                head = self.children[0].evaluate(state)
                tail = self.children[1:]
            return head, tail
    
    class xconcat_expr(expression):
        """
        Chain of expressions separated by a space (concatenation operator): x1 x2 x3 ...
        Values of subexpressions are converted to strings and concatenated WITHOUT space.
        This is an extension of Python syntax for concatenation of literal strings, like in:
               'Python' " is "  'cool'
        and is primarily intended for easy concatenation of variables to strings in HTML attr values construction:
               p style={"font-size:" size}
        """
        def evaluate(self, state, error = "expression to be string-concatenated evaluates to None"):
            return ''.join(STR(expr.evaluate(state), expr, error) for expr in self.children)
            # items = (STR(expr.evaluate(state)) for expr in self.children)
            # return ' '.join(item for item in items if item != '')     # empty strings '' silently removed from concatenation
        
    class simple_chain_expression(expression):
        """A chain built from the same binary operator: x OP y OP z OP ..."""
        oper = None                 # the operator function to be applied
        def evaluate(self, state):
            res = self.children[0].evaluate(state)
            for expr in self.children[1:]:
                val = expr.evaluate(state)
                res = self.oper(res, val)
            return res
    
    class xand_expr(simple_chain_expression):
        "chain of bitwise-and operators: &"
        oper = operator.and_
        name = '&'
    class xxor_expr(simple_chain_expression):
        "chain of bitwise-xor operators: ^"
        oper = operator.xor
        name = '^'
    class xor_expr(simple_chain_expression):
        "chain of bitwise-or (filtering) operators: |"
        name = '|'
        @staticmethod
        def oper(x, y):
            "x | y. If 'y' is a function or method, returns y(x) (filter application), else calculates x | y in a standard way."
            #if isfunction(y): return y(x)
            return x | y                            # here, 'y' can be a Filter instance
    class xpipeline(simple_chain_expression):
        """
        Chain of "pipeline" operators, x:fun(a,b) gets executed as fun(x,a,b).
        For oper() execution, fun(a,b) is represented as a partial function returned by our custom partial() implementation.
        """
        name = ':'
        @staticmethod
        def oper(x, partial_fun):
            return partial_fun(x)
            
            
    class xcomparison(chain_expression):
        "chain of comparison operators: < > == >= <= != in is, not in, is not"
        raise_null = False
        
    class xnot_test(expression):
        """not not not ..."""
        def evaluate(self, state):
            assert len(self.children) >= 2 and all(c.type == 'not' for c in self.children[:-1])
            neg = not (len(self.children) % 2)              # check parity of 'children' to see if negation appears even or odd no. of times
            val = self.children[-1].evaluate(state)
            return not val if neg else val
    
    class xand_test(simple_chain_expression):
        "chain of logical 'and' operators. Lazy evaluation: if false item is encountered, it's returned without evaluation of subsequent items"
        name = 'and'
        def evaluate(self, state):
            res = self.children[0].evaluate(state)
            for expr in self.children[1:]:
                if not res: return res
                res = res and expr.evaluate(state)
            return res
    class xor_test(simple_chain_expression):
        "chain of logical 'or' operators. Lazy evaluation: if true item is encountered, it's returned without evaluation of subsequent items"
        name = 'or'
        def evaluate(self, state):
            res = self.children[0].evaluate(state)
            for expr in self.children[1:]:
                if res: return res
                res = res or expr.evaluate(state)
            return res
    
    class xifelse_test(expression):
        """
        ... if ... else ... Lazy evaluation of arguments: only the true branch of the condition undergoes evaluation.
        "else" branch is optional, "else None" is assumed if "else" is missing.
        """
        def evaluate(self, state):
            assert len(self.children) in (2, 3)             # the expression is compactified, that's why a single child is not possible here
            if self.children[1].evaluate(state):
                return self.children[0].evaluate(state)
            if len(self.children) == 3:
                return self.children[2].evaluate(state)
            return None                                     # default None when "else..." branch is missing
    
    # class xempty_test(expression):
    #     """
    #     Test for emptiness: X OP [TEST]
    #     If the 1st operand (X) evaluates to false, a predefined default value is returned:
    #     - '' if operator OP is ?
    #     - None if operator OP is !
    #     Otherwise, X is returned unmodified. If an optional 2nd operand (TEST) is present and evaluates to false,
    #     the default ('' or None) is returned regardless of the value of X.
    #     Additionally, if OP = ?, any exceptions raised during evaluation of X are caught and
    #     treated the same as if X was false.
    #     """
    

    ###  EXPRESSIONS - COLLECTIONS  ###

    class xlist(expression):
        def evaluate(self, state):
            return [child.evaluate(state) for child in self.children]

    class xtuple(expression):
        def evaluate(self, state):
            return tuple(child.evaluate(state) for child in self.children)

    class xexpr_tuple(xtuple):
        pass
        
    class xset(expression):
        def evaluate(self, state):
            return set(child.evaluate(state) for child in self.children)

    class xdict(expression):
        def evaluate(self, state):
            items = []
            assert len(self.children) % 2 == 0          # there's always an even no. of children after reduction of dict_pair
            for i in range(0, len(self.children), 2):
                key_child, val_child = self.children[i:i+2]
                items.append((key_child.evaluate(state), val_child.evaluate(state)))
            return dict(items)

    class xstring_format(expression):
        def evaluate(self, state):
            return self._render_all(self.children, state)       # renders embedded expressions, in addition to static text
            

    ###  EXPRESSIONS - LITERALS  ###

    class literal(expression):
        isstatic = True
        ispure   = True
        value    = None
        def setup(self):            self.value = self.text()
        def analyse(self, ctx):     pass
        def evaluate(self, state):  return self.value
    
    class xstring_raw(literal):
        def setup(self):
            self.value = self.text()[2:-1]          # remove surrounding quotes: '' or "", and the leading "r" indicating a raw string

    class xnumber(literal):
        def setup(self):
            s = self.text()
            try:
                self.value = int(s)
                return
            except: pass
            self.value = float(s)
    
    class xboolean(literal):
        def setup(self):
            self.value = (self.text() == 'True')

    class xnone(literal):
        def setup(self):
            self.value = None

    class xattr_short_lit(literal): pass


    ###  STATIC nodes  ###
    
    class xpath_import(static): pass
    class xsymbol(static):      pass
    class xname_id(static):     pass
    class xname_xml(static):    pass

    class xtext_quot(static):
        r"""In `text_quot1` and `text_quot2` productions, Python's escape characters (\n \t \x.. etc.) must be decoded."""
        def setup(self):
            self.value = decode_escapes(self.text())
        
    class xtext_quot1(xtext_quot): pass
    class xtext_quot2(xtext_quot): pass

    class xnl(static):          pass
    class xtext(static):        pass
    class xdedent(static):      pass
    class xappend(static):      pass
    class xmargin(static):      pass
    
    class xmargin_out(static):
        """
        Vertical margin composed of 1+ newlines that preceeds an outlined block.
        A trailing newline (\n) is truncated from its `value` and moved out to a subsequent sibling node
        as a leading \n to mark that that node should be rendered in "outline" rather than "inline" mode.
        Every <margin_out> IS followed by a node (block) by grammar rules. The transition of the singleton
        newline is performed in xblock.translate().
        """
        def setup(self):
            self.value = self.text()[:-1]

    class xescape(static):
        def setup(self):
            escape = self.text()
            assert len(escape) == 2 and escape[0] == escape[1]
            self.value = escape[0]                          # the duplicated char is dropped
    
    class xup_indent(static):
        """Marks occurrence of an extra 1-space indentation in a title line. Renders to empty string."""
        def render(self, state):
            return ''
    
    class indent(node):
        whitechar = None
        def translate(self, state):
            """Called when INDENT/DEDENT surround a block."""
            state.indent(self.whitechar)
            return None
        def render(self, state):
            """Called when INDENT/DEDENT surround a line within a text block."""
            state.indent(self.whitechar)
            return ''

    class dedent(node):
        whitechar = None
        def translate(self, state):
            state.dedent(self.whitechar)
            return None
        def render(self, state):
            state.dedent(self.whitechar)
            return ''

    class xindent_s(indent):
        whitechar = ' '
    class xindent_t(indent):
        whitechar = '\t'
    class xdedent_s(dedent):
        whitechar = ' '
    class xdedent_t(dedent):
        whitechar = '\t'


    # ###  SYNTHETIC nodes  ###
    #
    # class merged(static):
    #     """
    #     An artificial node created during compactification by merging several sibling nodes that are all pure (or static, in particular).
    #     Values of the original nodes (strings to be concatenated) are retrieved from their render().
    #     """
    #     value = None        # pre-rendered output of the compactified nodes
    #     ex = None           # if FalseValueEx exception was caught during rendering, it's stored here as an (exception, traceback) pair
    #
    #     def __init__(self, node, state):
    #         self.tree = node.tree
    #         self.fulltext = node.fulltext
    #         self.pos = node.pos
    #         try:
    #             self.value = node.render(state)
    #         except FalseValueEx as ex:
    #             self.ex = (ex, sys.exc_info()[2])
    #
    #     def merge(self, node, state, sep):
    #         self.pos = (self.pos[0], node.pos[1])
    #         if self.ex: return                          # we already know that an exception will be raised upon self.render(), no need to append new nodes
    #         try:
    #             nodeValue = node.render(state)
    #             self.value += sep + nodeValue
    #         except FalseValueEx as ex:
    #             self.ex = (ex, sys.exc_info()[2])
    #
    #     def render(self, state):
    #         if self.ex: reraise(None, self.ex[0], self.ex[1])
    #         return self.value
    #
    #     def info(self):
    #         from nifty.util import escape as slash_escape
    #         return "%s at position %s rendering: %s" % (self.infoName(), self.pos, slash_escape(str(self.value)))
    #
    # ###  UTILITY METHODS  ###
    #
    # @staticmethod
    # def _compactify_siblings_(nodes, state, sep = u''):
    #     "Compactify a list of sibling nodes, by compactifying each one separately when possible and then merging neighboring static nodes."
    #     out = []
    #     last = None         # the current last <merged> node; can be expanded if the subsequent node is also pure
    #
    #     for node in nodes:
    #         #print(' ', node, node.check_pure())
    #         if node.check_pure():                               # a pure node that can be reduced into a <merged> node?
    #             if last: last.merge(node, state, sep)
    #             else:
    #                 last = NODES.merged(node, state)
    #                 out.append(last)
    #         else:                                               # non-pure node? let's compactify recursively its subtree and append
    #             node.compactify(state)
    #             out.append(node)
    #             last = None
    #
    #     return out
    

#####################################################################################################################################################
#####
#####  HypertagAST
#####

class HypertagAST(BaseTree):

    NODES  = NODES              # must tell the BaseTree's rewriting routine where node classes can be found
    parser = None               # instance of Grammar to be used by super class __init__() and parse() to convert input text to AST
    _use_init = False

    ###  Configuration of rewriting process  ###
    
    # nodes that will be ignored during rewriting (pruned from the tree)
    _ignore_  = "nl ws space gap comma verbatim inline_comment " \
                "mark_struct mark_verbat mark_normal mark_markup mark_embed mark_eval mark_def mark_comment"
    
    # nodes that will be replaced with a list of their children
    _reduce_  = "modifier block_control target core_blocks tail_blocks headline body_text generic_control generic_struct " \
                "item_import rename try_long try_short special_tag head_verbat head_normal head_markup " \
                "tail_for tail_if tail_verbat tail_normal tail_markup core_verbat core_normal core_markup " \
                "attrs_def attrs_val attr_val value_of_attr args arg " \
                "embedding embedding_braces embedding_eval target " \
                "expr_root subexpr slice subscript trailer trailer_filt atom literal dict_pair " \
                "string string_quot1 string_quot2"
    
    # nodes that will be replaced with their child if there is exactly 1 child AFTER rewriting of all children;
    # they must have a corresponding x... node class, because pruning is done after rewriting, not before
    _compact_ = "factor_filt factor_strict factor_var factor pow_expr term arith_expr shift_expr and_expr xor_expr or_expr concat_expr " \
                "pipeline comparison not_test and_test or_test ifelse_test expr_tuple " \
                "block_struct"

    _reduce_anonym_ = True      # reduce all anonymous nodes, i.e., nodes generated by unnamed expressions, typically groupings (...)
    _reduce_string_ = True      # if a node to be reduced has no children but matched a non-empty part of text, it shall be replaced with a 'string' node
    

    ###  Environment  ###
    
    module   = None             # the module this script was imported from, as runtime.Module instance; can be None (script from a string)
    filename = None             # name of the file this script comes from; for error messages
    runtime  = None             # instance of Runtime that loaded this document and controls how external modules and symbols are imported
    context  = None             # dict of symbols passed to self.translate() as `context`
    
    ###  Output of parsing and analysis  ###

    text    = None              # full text of the input string fed to the parser
    ast     = None              # raw AST as generated by Pasimonious; for read access
    root    = None              # root node of the final tree after rewriting

    # symbols   = None            # dict of all top-level symbols as name->node pairs
    # hypertags = None            # dict of top-level hypertags indexed by name, for use by the client as hypertag functions
    #                             # includes imported hypertags (!), but not external ones, only the native ones

    
    def __init__(self, script, module, verbose = False):

        self.module   = module
        self.runtime  = module.runtime
        self.filename = module.filename
        self.parser   = Grammar.get_parser(script)
        
        # replace indentation with special characters INDENT/DEDENT
        flat_script = self.parser.preprocess(script, verbose = verbose)
        
        # parse input text to the 1st version of AST (self.ast) as returned by Parsimonious,
        # then rewrite it to custom NODES.* classes rooted at self.root
        try:
            super(HypertagAST, self).__init__(flat_script)
            
        except IncompleteParseError as ex:
            L = 15
            line, line_num = self._locate_error(script)
            dots = "..." if len(line) > L else ""
            raise SyntaxErrorEx('invalid syntax near line %s: %s%s' % (line_num, line[:L], dots))
        
        if self.root is None:           # workaround for Parsimonious bug in the special case of empty document (Parsimonious returns None instead of a tree root)
            self.root = NODES.xdocument(self, ObjDict(start = 0, end = 0, children = [], expr_name = 'document'))

        assert isinstance(self.root, NODES.xdocument)


    def _locate_error(self, script):
        """
        Find the first line of `script` that causes Parsimonious' IncompleteParseError. Uses binary search and script truncation.
        
        This method relies on the fact that truncating an arbitrary number of trailing lines of a script
        never introduces syntax errors unless they already have been present in the preceeding lines.
        For this to be true, all control blocks are allowed by Hypertag grammar to have empty body.
        """
        lines = script.split('\n')
        
        line_good = 0               # the last known line before the error
        line_bad  = len(lines)      # the first known line after or containing the error
        
        while line_good + 1 < line_bad:
            split   = (line_good + line_bad) // 2
            partial = '\n'.join(lines[:split])
            flat    = self.parser.preprocess(partial)
            try:
                super(HypertagAST, self).__init__(flat)
                line_good = split
                
            except IncompleteParseError:
                line_bad = split

        if not line_bad: line_bad = 1
        return lines[line_bad-1].strip(), line_bad
        
    @staticmethod
    def make_context(tags, variables):
        """
        Combine two dicts of symbols into one. If a symbol lacks a leading mark ($,%), the mark is imputed
        as $ (for `variables`), or % (for `tags`).
        """
        symbols = {}

        if tags:
            # TODO: check if names of tags are non-empty and syntactically correct
            symbols.update({name if name[0] in (MARK_TAG, MARK_VAR) else TAG(name) : value for name, value in tags.items()})
        if variables:
            symbols.update({name if name[0] in (MARK_TAG, MARK_VAR) else VAR(name) : value for name, value in variables.items()})
        
        return symbols

        
    def analyse(self, builtins = None):
        "Link occurences of variables and hypertags with their definition nodes, collect all symbols defined in the document."
        
        ctx = Context()
        self.root.predefine(builtins)
        self.root.analyse(ctx)
        
    def translate(self, __builtins__ = None, __tags__ = None, **variables):
        
        if __builtins__ is None:
            __builtins__ = self.runtime.import_builtins()
        self.context = self.make_context(__tags__, variables)
        
        self.analyse(__builtins__)
        
        dom, symbols, state = self.root.translate(State())        # calls NODES.xdocument.translate(), see there for description of returned objects
        assert isinstance(dom, DOM.Root)
        # print('top-level symbols: {symbols}')
        
        return dom, symbols, state

    def render(self):
        dom, symbols, state = self.translate()
        return dom.render()
        # output = dom.render()
        # if not output: return output
        # assert output[0] == '\n'        # extra empty line was prepended by Grammar.preprocess() and must be removed now
        # return output[1:]

    # def __getitem__(self, tag_name):
    #     """Returns a top-level hypertag node wrapped up in Hypertag, for isolated rendering. Analysis must have been performed first."""
    #     # TODO
    #     return self.hypertags[tag_name]
        




import os, sys, importlib

from hypertag.core.errors import ImportErrorEx, ModuleNotFoundEx
from hypertag.core.grammar import MARK_VAR, MARK_TAG, VAR, TAG, TAGS
from hypertag.core.ast import HypertagAST

PATH_SEP = os.path.sep


#####################################################################################################################################################
#####
#####  UTILITIES
#####

# set Python's global recursion limit to at least MIN_RECURSION_LIMIT to avoid RecursionError
# when parsing deeply nested scripts
MIN_RECURSION_LIMIT = 20000
sys.setrecursionlimit(max(sys.getrecursionlimit(), MIN_RECURSION_LIMIT))


def _join_path(base, path, sep = '.', ext = None):
    """
    Convert a relative import `path` to an absolute one by appending it to an absolute `base` path.
    The `base` can be any type of path: a file path (slash-separated) or an import path (dot-separated) -
    the `sep` argument determines what character is treated as a separator; the same character is inserted
    in the result instead of dots, in the `path` part of the output string. If `ext` is provided,
    it is appended as a file extension of the output path.
    """
    if base is None: return None
    
    orig_base = base
    orig_path = path
    
    assert path[:1] == '.'                  # `path` is a relative import path
    path = path[1:]
    
    if base[-1:] == sep:                    # truncate a trailing separator in the base path
        base = base[:-1]

    while path[:1] == '.':                  # move upwards the package tree if `path` starts with multiple dots
        if sep not in base:
            raise ImportErrorEx("attempted relative import (%s) beyond top-level package or folder (%s)" % (orig_path, orig_base))
        base = base.rsplit(sep, 1)[0]
        path = path[1:]

    fullpath = base + sep + path.replace('.', sep)
    if ext: fullpath += '.' + ext
    
    return fullpath


#####################################################################################################################################################
#####
#####  MODULES
#####

class Module:
    """Wrapper around all types of imported modules (Hypertag's, Python's)."""
    
    runtime  = None     # Runtime that created this module
    location = None     # canonical path of this module, for deduplication and caching; there can be many
                        # non-canonical (e.g., relative) paths pointing to the same module, but only one canonical path
    filename = None     # name of the file if this module was loaded from disk; mapped to __file__ inside a script
    package  = None     # Python package path of this module's package, if available; mapped to __package__ inside a script
    symbols  = None     # cached dict of this module's symbols; each symbol has a leading mark % or $
    
    def __init__(self, **kwattrs):
        for attr, value in kwattrs.items():
            setattr(self, attr, value)
    
    def __getitem__(self, key):
        return self.symbols[key]
        
    def get(self, key, default = None):
        return self.symbols.get(key, default)
        

class PyModule(Module):
    """Wrapper around a regular Python module and its global symbols."""
    
    module = None       # the original Python module
    
    def __init__(self, module):
        """
        Pull symbols (tags & variables) from a Python module and return as a dict.
        All top-level symbols are treated as variables; tags are pulled from a special module-level dictionary `__tags__`.
        """
        self.module   = module
        self.location = getattr(module, '__name__', None)
        self.filename = getattr(module, '__file__', None)
        self.package  = getattr(module, '__package__', None)
        
        self.symbols  = {VAR(name) : getattr(module, name) for name in dir(module)}
        
        # retrieve __tags__
        tags = self.symbols.pop(VAR('__tags__'), None)
        if tags:
            if not isinstance(tags, dict): raise ImportErrorEx("__tags__ must be a dict in %s" % module)
            
            # make sure that the tags being imported are assigned to the same names as configured in their Tag.name property
            for name, tag in tags.items():
                if name != tag.name: raise ImportErrorEx("tag's internal name (%s) differs from its public name (%s) in %s" % (tag.name, name, module))
            
            self.symbols.update({name if name[0] == MARK_TAG else TAG(name) : tag for name, tag in tags.items()})

class HyModule(Module):
    """Wrapper around a Hypertag script to be executed (translated). Stores results of translation."""
    
    script = None       # plain text of the Hypertag script
    dom    = None       # output DOM produced by translation
    state  = None       # internal State at the end of translation; needed when hypertags from this module are to be expanded
    
    def translate(self, builtins, __tags__, **variables):
    
        ast = HypertagAST(self.script, self)
        self.dom, self.symbols, self.state = ast.translate(builtins, __tags__, **variables)
        return self
        
    
class RootModule(HyModule):
    """Wrapper module for a top-level script, to provide a referrer module for other (imported) scripts."""
    

#####################################################################################################################################################
#####
#####  LOADERS
#####

class Loader:
    """
    Finding, loading, and caching modules.
    A loader takes an import path as specified in `import` block of a Hypertag script,
    together with the location of the referrer module; converts the path to its canonical form (subclass-dependent);
    loads the module and returns it as a dict of symbols. Caching of modules can be performed internally.
    Each Loader subclass represents a specific naming convention and an ontology of valid import paths.
    """
    cache = None        # dict of cached modules indexed by their canonical path (module.location);
                        # each Loader subclass is responsible for managing the cache by itself
    
    def __init__(self):
        self.cache = {}
    
    def load(self, path, referrer, runtime):
        """
        Try to load a module given its (non-canonical) path and a referrer module, and return a Module instance.
        If the path is invalid or the module can't be found, None is returned.
        If this method is overriden by subclasses, `self.cache` should be checked to avoid repeated loading
        of the same module every time after a `path` is converted to a canonical `location`; every newly loaded module
        should be put into cache.
        """
        location = self._find(path, referrer)
        if location is None: return None
        if location in self.cache: return self.cache[location]
        
        module = self.cache[location] = self._read(location, runtime)
        return module

    def _find(self, path, referrer):
        """
        Convert a path to its canonical form. The `referrer` is the module where the path occured.
        Return None if the path is invalid for this type of modules.
        """
        return None

    def _read(self, location, runtime):
        """Read a module identified by `location`. Return None if the module is missing."""
        return None
    
    @staticmethod
    def _make_absolute(path, referrer):
        """Utility method for internal purposes: converting a python import path to an absolute path."""
        if path[:1] != '.': return path                 # `path` is absolute, return it unchanged
        return _join_path(referrer.package, path)       # `path` is relative, convert it to an absolute one
        
        
class PyLoader(Loader):
    """
    Loader of regular Python modules. They are identified by absolute package paths (dot-separated, no initial dot),
    which are used as modules' unique locations. When importing a Python module from a Hypertag script using
    a RELATIVE package path, the script's module must have its `package` property set, otherwise it will be impossible
    to infer the referred module's package path.
    """

    def _find(self, path, referrer):
        """Absolute and relative Python paths are supported. The latter require that referrer's package is set."""
        return self._make_absolute(path, referrer)
        
    def _read(self, location, runtime):
        try:
            module = importlib.import_module(location)
            return PyModule(module)
        
        except ModuleNotFoundError:
            return None


class HyLoader(Loader):
    """
    Loader of Hypertag scripts located on disk.
    Import paths have the same structure as in Python: with a dot '.' separating subfolders (packages).
    Location of a module is defined as an absolute file system path to the script file.

    A relative import path (.XYZ) is converted to a file path rooted at the referrer's module folder
    (referrer.filename must be present) to obtain the location.
    
    For absolute paths, different methods are tried to obtain the location. The first one that succeeds
    (i.e., the file exists) is selected:
    
    1. resolve(import_path, referrer) is called to get a location; `resolve` function is an optional argument
       of HyLoader instantiation;
    2. location is being searched for using Python's import mechanism, like if the script was a Python file - this only
       works when the import path contains a package specifier, which points to a valid Python package (with __init__.py);
    3. location is created as a directory path based at the current folder of the process;
    4. location is created as a directory path based at the referrer's folder (like for relative imports).
    
    Import paths CAN refer to folders which are NOT valid Python packages (don't have __init__.py inside).
    HyLoader does not check the existence of __init__.py on directory paths to scripts.
    """
    
    SCRIPT_EXTENSION = 'hy'         # default file extension of Hypertag scripts

    def __init__(self, resolve = None):
        super(HyLoader, self).__init__()
        self.resolve = resolve

    def load(self, path, referrer, runtime):
        
        # whenever possible, python package path (python_path) of the module is inferred, so that its `package` can be set
        location = python_path = None
        ref_root = os.path.dirname(referrer.filename) if referrer.filename else None
        
        # relative import path is always resolved relative to the referrer's folder
        if path[:1] == '.':
            if ref_root:
                location = self._join_path(ref_root, path)
                python_path = self._make_absolute(path, referrer)
                
        # absolute import path can be resolved in several different ways; the first one that returns a valid file path is used
        else:
            # 1. try calling a resolve() function
            if self.resolve:
                location = self.resolve(path, referrer)
                if location in self.cache: return self.cache[location]
                if location and not os.path.exists(location):
                    location = None

            # 2. try the Python's standard import mechanism (importlib); must be applied to the parent package, not the script itself
            if not location and '.' in path:
                package_name, filename = path.rsplit('.', 1)
                pkg = importlib.import_module(package_name)                 # package_name is an absolute path, so referrer is not needed here
                package_path = pkg.__file__
                if package_path.endswith('.py'):
                    package_path = os.path.dirname(package_path)            # truncate /__init__.py part of a package file path
                
                location = package_path + PATH_SEP + filename + '.' + self.SCRIPT_EXTENSION
                python_path = package_path + '.' + filename
                if location and not os.path.exists(location):
                    location = python_path = None
                
            # 3. try using the process' current folder as a root
            if not location:
                location = self._join_path(os.getcwd(), '.' + path)
                if location in self.cache: return self.cache[location]
                if location and not os.path.exists(location):
                    location = None

            # 4. try using the referrer's folder as a root
            if not location and ref_root:
                location = self._join_path(ref_root, '.' + path)
                python_path = self._make_absolute('.' + path, referrer)

        if not location or not os.path.exists(location):
            return None
        
        package = python_path.rsplit('.', 1)[0] if python_path else None
        
        script = open(location).read()
        
        module = self.cache[location] = runtime.translate(script, location, package)
        module.location = location
        module.package = package
        
        return module
        
    
    def _join_path(self, root, path):
        """Convert a python import `path` (with dots) to a file path (with slashes) and append to a root folder's path."""
        
        return _join_path(root, path, PATH_SEP, self.SCRIPT_EXTENSION)
        

#####################################################################################################################################################
#####
#####  RUNTIME
#####

class Runtime:
    """
    Runtime execution environment of Hypertag scripts.
    Defines a set of built-in symbols, a target language, and an escape function.
    Manages modules loading through a collection of predefined loaders.
    Modules are identified by "paths". The meaning of a particular path is determined by its loader.
    """

    # list of modules to be imported automatically into a script upon startup of translation (built-in symbols)
    BUILTINS = ['builtins', 'hypertag.builtins']
    
    # cached dict of built-in symbols, to avoid recalculation for every new AST
    _builtins = None

    language = None     # target language the documents will be compiled into, defined in subclasses
    #compact  = True    # if True, compactification is performed after analysis: pure (static, constant) nodes are replaced with their pre-computed render() values,
                        # which are returned on all subsequent render() requests; this improves performance when
                        # a document contains many static parts and variables occur rarely
    escape   = None     # escaping function or static method that converts plaintext to target language; typically, when assigned
                        # in a subclass, staticmethod() must be applied as a wrapper to prevent this attr be treated as a regular method:
                        #   escape = staticmethod(custom_function)

    loaders  = None     # list of Module subclasses whose static load() is called in sequence to find the first one
                        # that is able to locate and load a module by a given path
    
    default_loaders = [HyLoader, PyLoader]      # loaders to be used when no others are passed to __init__
    
    
    def __init__(self, loaders = None):
        self.loaders = loaders or self.default_loaders
        self.loaders = [loader if isinstance(loader, Loader) else loader() for loader in self.loaders]

    def import_builtins(self):
        """
        Import default symbols that shall be available to every script upon startup.
        This typically means all general-purpose symbols + standard tags/variables specific for a target language.
        """
        if self._builtins is None:
            self._builtins = {}
            for path in self.BUILTINS:
                module = self.import_module(path, RootModule(), None)
                self._builtins.update(module.symbols)
                
        return self._builtins
        
    def import_module(self, path, referrer, ast_node):
        """Import symbols that are defined in a module identified by `path`. Return as an instance of Module."""
        
        assert path

        for loader in self.loaders:
            module = loader.load(path, referrer, self)
            if module: return module

        raise ModuleNotFoundEx("import path not found '%s', try setting __package__ or __file__ when calling render()" % path, ast_node)
        
    def translate(self, __script__, __file__ = None, __package__ = None, __tags__ = None, **variables):
        
        builtins = self.import_builtins()
        builtins[VAR('__file__')]    = __file__
        builtins[VAR('__package__')] = __package__
        
        module = HyModule(runtime = self, script = __script__, filename = __file__, package  = __package__)
        module.translate(builtins, __tags__, **variables)
        
        return module
        
    def render(self, __script__, __file__ = None, __package__ = None, __tags__ = None, **variables):
        
        module = self.translate(__script__, __file__, __package__, __tags__, **variables)
        return module.dom.render()
        




import re, itertools
from hypertag.nifty.util import unique as unique_list, merge_spaces
from hypertag.std.registry import Registry


register = Registry()


########################################################################################################################################################
#####
#####  GENERAL PURPOSE tags
#####

_re_dedent = re.compile(r'(?m)^\s+')
_re_inline = re.compile(r'\s+')


@register.tag
@register.var
def dedent(text):
    """Remove all line indentation in `text`. The indentation may differ between lines and it still gets fully removed."""
    return _re_dedent.sub('', text)
    # if full: return _re_indent.sub('', text)
    # return del_indent(text)

@register.tag
@register.var
def unique(text, strip = True):
    """
    Remove duplicate lines in `text`. The order of remaining lines is preserved.
    If strip=True (default), the lines are stripped of leading & trailing whitespace before comparison,
    and all empty lines are removed.
    """
    lines = text.splitlines()
    if strip: lines = list(filter(None, map(str.strip, lines)))
    uniq = unique_list(lines)
    return '\n'.join(uniq)

@register.tag
@register.var
def inline(text):
    """
    Strip leading/trailing whitespace in `text`, replace newlines and tabs with spaces, merge multiple adjacent spaces.
    Similar to normalize-space() in XPath.
    """
    return _re_inline.sub(' ', text).strip()

@register.tag
@register.var
def lower(text):
    """A tag/function/filter that converts `text` to lowercase. Inner tag names and attributes are converted, too (!)."""
    return text.lower()

@register.tag
@register.var
def upper(text):
    """A tag/function/filter that converts `text` to uppercase. Inner tag names and attributes are converted, too (!)."""
    return text.upper()

@register.var
def cycle(*sequences, **params):   #stop = 'first'
    """
    A generator that cycles through each of the provided iterables IN PARALLEL and yields tuples containing
    one element from each iterable - like zip(), but the cycling runs indefinitely (if stop=False), or until
    the FIRST iterable gets exhausted (if stop='first', default), or until the LONGEST iterable gets exhausted
    (stop='longest'). When an iterable is exhausted along the way, its sequence of elements is repeated.
    This function is inspired by Django's {%cycle%} tag, yet it differs from this tag in that it doesn't require
    maintaining a hidden state, which enhances code readability.
    This function can be used, for instance, to alternate colors of rows in a table.
    """
    stop = params.pop('stop', 'first')
    if params: raise Exception('unrecognized keyword argument "%s" in cycle()' % list(params.keys())[0])
    
    if not sequences: return []
    length = None
    
    if stop == 'first':
        sequences = list(sequences)
        sequences[0] = list(sequences[0])               # convert the 1st iterable to a list, so that we can check its length
        length = len(sequences[0])                      # length of the output sequence
        
    elif stop == 'longest':
        sequences = list(map(list, sequences))          # convert each sequence to a list, so that we can check their lengths
        length = max(map(len, sequences))               # ...and find the longest one
    
    cycles = map(itertools.cycle, sequences)            # create indefinite-loop iterators over each iterable
    
    return itertools.islice(zip(*cycles), length)

@register.var
def changes(seq, first = True):
    """
    Iterates over a given sequence and yields pairs (value, change), where `value` is the next element of `seq`,
    and `change` is True if the value differs from the previous one, False otherwise.
    In the first tuple, `change` equals `first` (True by default).
    This generator can be used in loops as a replacement for Django's {%ifchanged%} tag.
    """
    prev = None
    for i, value in enumerate(seq):
        if i == 0:
            yield value, first
        else:
            yield value, (value != prev)
        prev = value


########################################################################################################################################################

"""
Terminal tags:
- resource            -- terminal tag that marks a given block as a resource that shall be moved to HTML's <meta> section
- error               -- inserts a standard error message in a place of occurrence; root document node might collect all <error> nodes and produce a combined (hidden) error message

Functions inspired by Django/Ninja:
- findchange(seq)            -- similar to enumerate() but outputs a binary flag whenever a value of the sequence changes relative to the previous one;
                                replacement for Django's {%ifchanged%}
  finddiff() changes()
  
- url ??

Filters:
- groupby(key, d=None)       -- when applied to a list of dicts, returns a list of groups of records having the same value of `key`;
                                each group is a list of records; `key` can be a tuple of key names, names can refer to related objects "x.y.z";
                                replacement for Django's {%regroup%}
"""




# -*- coding: utf-8 -*-
'''
COPIED FROM "nifty" TOOLBOX with MODIFICATIONS:  https://github.com/mwojnars/nifty/blob/master/parsing/parsing.py

LICENSE: MIT
'''

import copy
from .util import isstring, escape, flatten


########################################################################################################################################################
###
###  UTILITIES
###

class ParserError(Exception):
    
    MAXLEN = 20
    pos    = None       # (start,end) position where the error occured, or None
    line   = None       # derived from 'pos'
    column = None       # derived from 'pos'
    text   = None       # initial MAXLEN characters of the input fragment that caused the error
    node   = None

    def __init__(self, msg, node = None, cause = None):
        if node:
            self.node, self.pos, self.text = node, node.pos, node.text(self.MAXLEN)
        if self.pos:                                            # calculate the line number and column of self.pos, if possible
            prefix = node.fulltext[:self.pos[1]]
            self.line = prefix.count('\n') + 1
            self.column = len(prefix) - prefix.rfind('\n')
        msg = self.make_msg(msg) or msg
        if cause: msg += " because of %s: %s" % (type(cause).__name__, cause)
        super(ParserError, self).__init__(msg)
        
    def make_msg(self, msg):
        if self.pos:
            return msg + " at line %s, column %s (%s)" % (self.line, self.column, self.text)


########################################################################################################################################################
###
###  TREE
###

class Tree(object):
    """Semantic tree. Produced as a result of rewriting of the original raw AST generated by parser.
    In contrast to raw AST, which always comprises nodes of the same type, no matter what language is parsed,
    this Tree comprises custom domain-specific nodes which implement semantic analysis specific to a given language.
    That's why rewriting from AST to Tree is needed.
    Subclasses of Tree should contain inner node classes named after the names of non-terminals of the grammar, 
    precedeed by leading 'x' to differentiate node classes from Tree methods and attributes.
    Instead of inner classes, you can use methods.
    """
    parser  = None              # instance of AST parser class produced by parser generator; only one method is used: parse(text) -> AST
    #Context = None              # Context subclass that shall be used in semantic analysis

    _use_init = True            # only set to True in old code that uses init(...) with arguments; otherwise set False in subclasses and override setup() in nodes

    # the sets below can be given as a space-separated string of names instead of a set
    _ignore_  = set()           # set of non-terminals that should be ignored during rewriting (pruned from the tree)
    _reduce_  = set()           # set of non-terminals that should be replaced with a list of their children
    _compact_ = set()           # set of non-terminals that should be replaced with their child if there is exactly 1 child AFTER rewriting of all children

    text    = None              # full text of the input string fed to the parser
    ast     = None              # raw AST generated by the parser; for read access by the client
    root    = None              # root node of the final tree after rewriting
    
    def __init__(self, text = None, ast = None, stopAfter = None):
        """Build Tree, either from input 'text' (will be parsed to raw AST and then rewritten to Tree), 
        or from raw AST 'ast' (only rewriting will be done).
        """
        def _split(names):
            if isstring(names): names = names.split()
            if isinstance(names, list): names = set(names)
            return names

        self._ignore_  = _split(self._ignore_)
        self._reduce_  = _split(self._reduce_)
        self._compact_ = _split(self._compact_)
        
        self.text = text
        self.ast = self.parse(text) if text else ast                # parse input text to raw AST; keep AST for reference by the client
        if stopAfter == "parse": return
        
        self.root = self.rewrite(self.ast)                          # rewrite AST to a tree of <Tree.node> nodes
        if not self.root: return
        
        self.root._enrich()                                         # enrich nodes with basic semantic information
        
        # bottom-up initialization of nodes by calls to custom init(), no arguments
        if not self._use_init:
            self.root._setup_all()

    def __str__(self):
        "For printing of the tree. Walks the tree via node.children, collects info() lines and concatenates them with proper indentation."
        lines = self.info(self.root)
        return '\n'.join(flatten(lines))
    
    @staticmethod
    def info(node, depth = 0):
        prefix = '  ' * depth
        if isstring(node): return prefix + node
        return prefix + node.info(), [Tree.info(n, depth+1) for n in node.children]

    def parse(self, text):
        "Parses raw text (string) into a syntax tree built of custom node classes (self.Tree). Can be overridden in subclasses."
        return self.parser.parse(text)

    def rewrite(self, astroot):
        "Recursively convert a node of raw AST and all its subtree into appropriate subclasses of ``Tree.node``. Generator-specific."
        raise NotImplementedError()
    
    def _rewriteNode(self, astnode):
        "Rewrites contents of a single node to corresponding Tree.node attributes: pos, type, children. Generator-specific."
        raise NotImplementedError()
    
    def compile(self):                                                                                                          #@ReservedAssignment
        "Full compilation, with postprocessing. Can be overriden in subclasses, but usually overriding postprocess() or analyse() is enough."
        data = self.analyse()                                   # run top-down semantic analysis of the tree
        code = self.root.compile()                              # generate output code
        return self.postprocess(code, data)

    def analyse(self):
        ctx = self.root.analyse(self.Context())
        return ctx.data

    def postprocess(self, code, data):
        "Override in subclasses to perform custom postprocessing of the compiled code and possibly change the return type, e.g., avoid returning 'data'."
        return code, data
        
    ### Base node classes
    
    class node(object):
        """Generic recursive node: passes all str() and compile() requests down to children, then concatenates returned strings.
        Base class for all nodes in rewritten trees.
        """
        
        tree     = None
        astnode  = None                     # original AST node this node was created from
        fulltext = None                     # full text of the input string fed to the parser; see text()
        pos      = None                     # (start,end) positions of the substring represented by this node in 'fulltext'
        type     = None                     # name of non-terminal that produced this node, [str]   @ReservedAssignment
        display  = ""                       # string representation of this node for debugging purposes
        
        children = []                       # list of child nodes after rewriting
        
        # enriched information
        parent = None                       # parent node; None if root
        sibling_prev = None                 # older sibling (before this node) AFTER rewriting
        sibling_next = None                 # younger sibling (after this node) AFTER rewriting
        
        @property
        def line(self):
            """
            Line number in `fulltext` where this node's match begins.
            NOTE: this method scans all fulltext up to self.pos, which can be costly.
            """
            prefix = self.fulltext[:self.pos[0]]
            return prefix.count('\n') + 1

        @property
        def column(self):
            prefix = self.fulltext[:self.pos[0]]
            return len(prefix) - prefix.rfind('\n')

        def __init__(self, tree, astnode): 
            self.tree = tree
            self.astnode = astnode
            self.fulltext = tree.text
            self.pos, self.type, self.children = tree._rewriteNode(astnode)
            if tree._use_init:
                self.init(tree, astnode)
            
        def init(self, tree = None, astnode = None):
            """
            Subclasses should override this method (args can be dropped!) instead of __init__,
            so that standard initialization is still performed beforehand, without explicit super() call.
            DEPRECATED. In new code, override setup() instead.
            """
        def setup(self):
            """
            Like init(), but doesn't take arguments. Can be overriden in subclasses to perform
            custom initialization of a node after standard rewriting and enrichment.
            In new code, setup() should be overriden instead of init(...), with Tree._use_init = False.
            """
            
        def _setup_all(self):
            """Recursive bottom-up initialization of nodes with calls to custom setup() methods, no arguments (!)."""
            for child in self.children: child._setup_all()
            self.setup()

        def text(self, maxlen = None):
            """The substring of source text matched by this node, or its leading 'maxlen' characters if maxlen != None. 
            None if this node was derived from other nodes during tree post-processing."""
            if not self.pos: return None
            end = self.pos[1] if maxlen is None else min(self.pos[1], self.pos[0] + maxlen)
            ret = self.fulltext[ self.pos[0] : end ]
            if end < self.pos[1]: return ret + "..."
            return ret
    
        def info(self):
            #d = self.__dict__.copy(); del d['fulltext']
            ds = "" #" %s" % d
            return "%s%s at position %s matching: %s" % (self.infoName(), ds, self.pos, escape(str(self.text())))
        def infoName(self):
            # override in subclasses to customize info() printout
            return "<%s>" % (self.type or self.__class__.__name__)
        
        def __str__(self): 
            return self.display or ''.join(str(c) for c in self.children)
        
        def child(self, idx):
            "Returns child with a given index, or None if missing."
            return self.children[idx] if len(self.children) > idx else None
        
        def _enrich(self):
            """Recursively enrich nodes with basic semantic information. Called right after rewriting."""
            
            prev = None
            for child in self.children:
                child.parent = self
                child.sibling_prev = prev
                if prev: prev.sibling_next = child
                prev = child

            for child in self.children:
                child._enrich()

        def analyse(self, ctx):
            """Top-down semantic analysis of the tree. 'ctx' is the opening Context (pre-context) of the current node. 
            Returns closing context (post-context) of the node, which may be the same object 'ctx' (!), but doesn't have to.  
            Parent nodes shall NEVER assume that 'ctx' doesn't get modified when passed down to a child.
            If this object is going to be used again after passing to a child (even if it's only passing to another child!), 
            a copy should be passed instead of the original.
            """
            for c in self.children: c.analyse(ctx.copy())
            return ctx

        def compile(self): #@ReservedAssignment
            "Top-down compilation of the tree. By default, returns concatenation of strings compiled by children nodes."
            return ''.join(c.compile() for c in self.children)

    class virtual(node):
        """A node that's artificially created after the tree is parsed, 
        having no corresponding AST node nor a position in the source text."""
        def __init__(self, tree):
            "No 'astnode' argument, we only keep the tree reference."
            self.tree = tree
    
    class static(node):
        """Static string (or value of another type) that can be inferred already during parsing (node initialization) and then returned 
        in str() and compile(). By default, the value is made up in __init__ by merging strings from all children."""
        typecast = str              # type or method that will be applied to the parsed string to convert it to a value of appropriate type
        
        def __init__(self, tree, astnode, string = None):
            "Additional argument 'string' is accepted, not present in base 'node'. It's used as a node value if children produce empty string."
            self.tree = tree
            self.fulltext = tree.text
            self.pos, self.type, children = tree._rewriteNode(astnode) if astnode else (None, None, [])
            string = ''.join(str(c) for c in children) or string or ''
            self.value = self.typecast(string)
            if tree._use_init:
                self.init(tree, astnode)
        
        def __str__(self): return str(self.value)
        def compile(self): return self.value                                                        #@ReservedAssignment
    
    class const(node):
        "A node that always generates the same output: cls.value. However, in general, it can be parsed from different substrings of input text (!)."
        value = ""
        def compile(self): return self.value                                                        #@ReservedAssignment
    
    class string(node):
        """Artificial node created at some intermediate stage of tree post-processing that may or may not correspond
        to a specific node of the initial AST. Typically produced from partial compilation of some other Tree nodes.
        """
        value = None
        type  = "(string)"                                                                          #@ReservedAssignment
        
        def __init__(self, tree, value, pos = None):
            self.tree = tree
            self.fulltext = tree.text
            self.value = value
            self.pos = pos
        def info(self):
            return "<%s> %s" % (self.type, escape(self.value))
        
    
    class empty(static): pass


########################################################################################################################################################
###
###  PARSIMONIOUS tree
###

class ParsimoniousTree(Tree):

    NODES = None                # the container (module/class/...) where node classes "x***" can be found; if None, self is used instead

    _reduce_anonym_ = True      # shall we reduce anonymous nodes? these are the nodes generated by unnamed expressions, typically groupings (...);
                                # leaf nodes (static strings) will be removed entirely from the tree, i.e., ignored rather than reduced
    _reduce_string_ = True      # if a node to be reduce has no children but matched a non-empty part of the text, shall it be replaced with a string node
                                # instead of raising an exception? 

    class reduced(Tree.string):
        "Like Tree.string, but adapted to reducing Parsimonious AST nodes."
        type = "(reduced)"
        
        def __init__(self, tree, astnode):
            Tree.string.__init__(self, tree, astnode.text, (astnode.start, astnode.end))
        

    def rewrite(self, astnode):
        def flat(): return self._rewriteNode(astnode)[-1]       # flattening: return children instead of the node
            
        if astnode is None: return None
        name = astnode.expr_name
        if not name and self._reduce_anonym_: return flat()     # flatten nodes without names; leaf nodes (static strings) removed entirely
        if name in self._ignore_: return []                     # remove nodes listed in _ignore_, together with their subtrees
        if name in self._reduce_:                               # flatten nodes listed in _reduce_
            if astnode.children: return flat()
            if astnode.start == astnode.end != None: return []
            if self._reduce_string_: return self.reduced(astnode)
            raise ParserError("Trying to reduce a node that has no children but consumed a non-empty part of the input at position (%s,%s): %s" % 
                              (astnode.start, astnode.end, astnode))
        
        # find corresponding inner class of the tree and instantiate (this will recursively rewrite nodes down the tree)
        nodeclass = getattr(self.NODES or self, 'x' + name)
        node = nodeclass(self, astnode)
    
        if name in self._compact_:                              # flatten nodes listed in _compact_
            if len(node.children) == 1: return node.children[0]

        return node
    
    def _rewriteNode(self, astnode):
        children = flatten(self.rewrite(c) for c in astnode.children)
        return (astnode.start, astnode.end), astnode.expr_name, children        # pos, type, children
    

########################################################################################################################################################
###
###  CONTEXT of semantic analysis
###

class Context(object):
    """Base class for classes representing current context of semantic analysis in Tree.analyse(), 
    for top-down transfer of contextual data from parent to child nodes and for bottom-up collection of global data from all nodes. 
    Can be modified along the way, when passed through the tree, and copied with .copy() method 
    when local modifications must be performed. 
    """
    
    Data = None                             # class (or function that returns instantiated object) of self.data
    
    def __init__(self):
        self.data = self.Data() if self.Data else None          # global semantic information about the syntax tree, shared by all copies of Context
        
    def copy(self):
        "A safe - shallow - copy of Context. Global data are shared by all copies."
        return copy.copy(self)
    



from hypertag.core.runtime import Runtime
from hypertag.std.html import html_escape
import hypertag.html


########################################################################################################################################################
#####
#####  HTML runtime
#####


class HyperHTML(Runtime):
    
    language = 'HTML'
    escape   = staticmethod(html_escape)
    BUILTINS = Runtime.BUILTINS + ['hypertag.html']
    


"""
Django standard filters for use with Hypertag.
A complete list and documentation of filters:  https://docs.djangoproject.com/en/3.1/ref/templates/builtins/#built-in-filter-reference

All standard Django filters are loaded to this module's namespace, so that they can be imported into other files
and used as functions or Hypertag filters (with a pipeline operator), for example:

    from hypertag.django.filters import slugify
    from hypertag.django.filters import dictsort

OR, if imported in a Hypertag script:

    from hypertag.django.filters import $slugify

"""

import django as __django__
from django.conf import settings as __settings__
from django.template.defaultfilters import register as __register_default__
from django.contrib.humanize.templatetags.humanize import register as __register_humanize__

##
## for now, Django's default settings are always used in those filters that depend on them
##

__settings__.configure()
__django__.setup()


globals().update(__register_humanize__.filters)
globals().update(__register_default__.filters)

del globals()['__django__']
del globals()['__settings__']
del globals()['__register_default__']
del globals()['__register_humanize__']



from hypertag.nifty.parsing import ParserError

########################################################################################################################################################

class HError(ParserError):
    def make_msg(self, msg):
        if not self.pos: return
        
        # convert `line`, `column` to coordinates of the original script from before INDENT/DEDENT encoding
        if self.node and self.column is not None:
            prefix  = self.node.fulltext[:self.pos[0]]
            symbols = self.node.tree.parser.symbols
            indents = prefix.count(symbols['INDENT_S']) + prefix.count(symbols['INDENT_T'])
            dedents = prefix.count(symbols['DEDENT_S']) + prefix.count(symbols['DEDENT_T'])
            indent  = indents - dedents
            column  = indent + self.column
        else:
            column  = None
        
        line = self.line - 1 if self.line else self.line
        
        if self.node and self.node.tree.filename:
            return msg + " in '%s', line %s, column %s (%s)" % (self.node.tree.filename, line, column, self.text)
        else:
            return msg + " at line %s, column %s (%s)" % (line, column, self.text)


# class HBaseException(BaseException):
#     """
#     Base class for exceptions that play internal role during execution of a Hypertag AST.
#     Their occurrence does NOT signal an error, rather they are intended to be caught by Hypertag internal code
#     at a higher level in the tree.
#     """
#
# class HBreak(HBaseException):
#     pass


########################################################################################################################################################

class SyntaxErrorEx(HError, SyntaxError):               pass
class ValueErrorEx(HError, ValueError):                 pass
class TypeErrorEx(HError, TypeError):                   pass
class NameErrorEx(HError, NameError):                   pass
class UnboundLocalEx(HError, UnboundLocalError):        pass
class UndefinedTagEx(HError):                           pass
class NotATagEx(HError):                                pass
class NoneStringEx(HError):                             pass
class ImportErrorEx(HError, ImportError):               pass
class ModuleNotFoundEx(ImportErrorEx):                  pass
class ParseErrorEx(HError):                             pass

class FalseValueEx(HError):
    """
    Empty (false) value was returned by an expression marked with "!" (obligatory) qualifier.
    Typically, this exception is caught at a higher level with a "try" block.
    """
    
class VoidTagEx(HError):
    """Raised when non-empty body is passed to a void tag (i.e., a tag that doesn't accept body)."""
    



########################################################################################################################################################
###
###  Regex patterns for character sets allowed in XML identifiers, to be put inside [...] in a regex.
###  XML identifiers differ substantially from typical name patterns in other computer languages. Main differences:
###   1) national Unicode characters are allowed, specified by ranges of unicode point values
###   2) special characters are allowed:  ':' (colon) '.' (dot) '-' (minus)
###      Colon is allowed as the 1st character according to XML syntax spec., although such a name may be treated as malformed during semantic analysis.
###      Others (dot, minus), are allowed on further positions in the string, after the 1st character.
###  Specification: http://www.w3.org/TR/REC-xml/#NT-NameStartChar
###

# human-readable:  [:_A-Za-z] | [\u00C0-\u00D6] | [\u00D8-\u00F6] | [\u00F8-\u02FF] | [\u0370-\u037D] | [\u037F-\u1FFF] | [\u200C-\u200D] | [\u2070-\u218F] | [\u2C00-\u2FEF] | [\u3001-\uD7FF] | [\uF900-\uFDCF] | [\uFDF0-\uFFFD] | [\U00010000-\U000EFFFF]
XML_StartChar  =  r"_A-Za-z\u00C0-\u00D6\u00D8-\u00F6\u00F8-\u02FF\u0370-\u037D\u037F-\u1FFF\u200C-\u200D\u2070-\u218F\u2C00-\u2FEF\u3001-\uD7FF\uF900-\uFDCF\uFDF0-\uFFFD\U00010000-\U000EFFFF"

# human-readable:  XML_StartChar | [0-9.\u00B7-] | [\u0300-\u036F] | [\u203F-\u2040]
XML_Char       =  XML_StartChar + r"0-9\.\-" + r"\u00B7\u0300-\u036F\u203F-\u2040" + ":"

# XML_EndChar is like XML_Char, but ":" is NOT allowed, to avoid collision with a trailing ":" used in Hypertag blocks
XML_EndChar    =  XML_Char[:-1]




"""
Hypertag's template processing backend for Django.
"""

from django.template import Origin, TemplateDoesNotExist, TemplateSyntaxError
from django.template.backends.base import BaseEngine
from django.template.backends.utils import csrf_input_lazy, csrf_token_lazy

from hypertag import HyperHTML



class Hypertag(BaseEngine):
    """
    Hypertag's template processing backend for Django. Produces HTML5 documents.
    This class should be installed in TEMPLATES, in Django project's settings.py:

    TEMPLATES = [
        {
            'BACKEND': 'hypertag.django.backend.Hypertag',
            'DIRS': [],
            'APP_DIRS': True,
            'OPTIONS': {},
        },
        # ... other engines here ...
    ]
    """
    
    # Name of the subdirectory containing the templates for this engine
    # inside an installed application.
    app_dirname = 'hypertag'


    def __init__(self, params):
        # print('setting up Hypertag backend...')
        params  = params.copy()
        options = params.pop('OPTIONS').copy()
        super().__init__(params)
        
        # `options` are ignored currently
        self.engine = HyperHTML()  #**options

    def from_string(self, template_code):
        
        return Template(self.engine, template_code)

    def get_template(self, template_name):
        
        tried = []
        for template_file in self.iter_template_filenames(template_name):
            try:
                script = open(template_file, 'rt').read()
                return Template(self.engine, script, template_file)

            except FileNotFoundError:
                tried.append((
                    Origin(template_file, template_name, self),
                    'Source does not exist',
                ))

        raise TemplateDoesNotExist(template_name, tried = tried, backend = self)


class Template:

    def __init__(self, engine, view, filename = None):
        self.engine = engine
        self.view = view
        self.filename = filename

    def render(self, context=None, request=None):
        if context is None:
            context = {}
        if request is not None:
            context['request'] = request
            context['csrf_input'] = csrf_input_lazy(request)
            context['csrf_token'] = csrf_token_lazy(request)
        if self.filename and '__file__' not in context:
            context['__file__'] = self.filename
        
        return self.engine.render(self.view, **context)
    
    


"""
Classes that represent Hypertag's native DOM (Document Object Model) that is produced as an output
of translation of the AST. The DOM can subsequently undergo "rendering" to a document in a target language.

DOM manipulation methods:
- find(path): generate a sequence of all (sub)nodes that match a given path.
- remove(path): remove from the DOM and return all (sub)nodes that match a given path;

@author:  Marcin Wojnarski
"""

import re, itertools
from copy import copy
from types import GeneratorType

from hypertag.nifty.util import Object

from hypertag.core.errors import VoidTagEx, TypeErrorEx


########################################################################################################################################################
#####
#####  UTILITIES
#####


def add_indent(text, indent, re_start = re.compile(r'\n(?=.)')):        # re.compile(r'(?m)^(?=.)')
    """
    Append `indent` string at the beginning of each line of `text` excluding the 1st line (!).
    Empty lines (containing zero characters, not even a space) are left untouched!
    """
    if not indent: return text
    return re_start.sub('\n' + indent, text)
    # if not text: return text
    # return indent + text.replace('\n', '\n' + indent)
    
def del_indent(text, indent = None):
    """
    Remove `indent` string from the beginning of each line of `text`, wherever it is present as a line prefix.
    If indent=None, maximum common indentation (get_indent()) is truncated.
    """
    if indent is None: indent = get_indent(text)
    if text.startswith(indent): text = text[len(indent):]
    return text.replace('\n' + indent, '\n')

def get_indent(text):
    """
    Retrieve the longest indentation string fully composed of whitespace
    that is shared by ALL non-empty lines in `text`, including the 1st line (if it contains a non-whitespace).
    """
    lines = text.split('\n')
    lines = list(filter(None, [l if l.strip() else '' for l in lines]))          # filter out empty lines
    if not lines: return ''

    for i, column in enumerate(zip(*lines)):        # zip() only eats up as many characters as the shortest line
        if not column[0].isspace() or min(column) != max(column):
            return lines[0][:i]
    else:
        size = min(map(len, lines))
        return lines[0][:size]                      # when all lines are prefixes of each other take the shortest one
    

########################################################################################################################################################
#####
#####  DOCUMENT OBJECT MODEL (DOM)
#####

class DOM:
    """
    List of DOM.Nodes that comprise (a part of) a body of a parent Node, or was produced as an intermediate
    collection of nodes during DOM manipulation.
    Provides methods for traversing DOM (sub)trees and selecting nodes,
    as well as flattening and cleaning up the list during node construction.
    
    All DOM node classes: Node, Root, Text - are defined as inner classes of this one.
    """
    nodes = None
    
    def __init__(self, *nodes, **params):  #_strict = True
        _strict = params.pop('_strict', True)
        if params: raise TypeErrorEx('unrecognized keyword argument "%s" in DOM.__init__()' % list(params.keys())[0])
        self.nodes = self._flatten(nodes) if _strict else list(nodes)
        # self.set_nodes(nodes, _strict)
        
    # def set_nodes(self, nodes, strict = True):
    #     self.nodes = self._flatten(nodes) if strict else list(nodes)
    
    def __bool__(self):             return bool(self.nodes)
    def __len__(self):              return len(self.nodes)
    def __iter__(self):             return iter(self.nodes)
    def __getitem__(self, pos):
        
        if isinstance(pos, int):
            return self.nodes[pos]
        if isinstance(pos, (int, slice)):
            return DOM(self.nodes[pos], _strict = False)
        # if isinstance(pos, (str, Tag)):
        return self.select(pos)
    
    @staticmethod
    def node(*args, **kwargs):
        """
        Create a DOM consisting of a single DOM.Node instance initialized with given arguments.
        A shortcut for DOM(DOM.Node(...)).
        """
        return DOM(DOM.Node(*args, **kwargs))
    
    @staticmethod
    def text(*args, **kwargs):
        """
        Create a DOM consisting of a single DOM.Text instance initialized with given arguments.
        A shortcut for DOM(DOM.Text(...)).
        """
        return DOM(DOM.Text(*args, **kwargs))
    
    @staticmethod
    def _flatten(nodes):
        """Flatten nested lists of nodes by concatenating them into the top-level list; drop None's."""
        result = []
        for n in nodes:
            if n is None: continue
            if isinstance(n, (list, DOM, GeneratorType)):
                result += DOM._flatten(n)
            elif isinstance(n, DOM.Node):
                result.append(n)
            else:
                raise TypeErrorEx("found %s in a DOM, expected DOM.Node" % type(n))
        return result
        
    def set_indent(self, indent):
        for n in self.nodes:
            n.set_indent(indent)
            
    def render(self):
        return ''.join(node.render() for node in self.nodes)

    def tree(self, indent = '', step = '  '):
        r"""Return a multiline \n-terminated string that presents this DOM's structure as a list of trees."""
        return ''.join(node.tree(indent, step) for node in self.nodes)
        
    def copy(self):
        """
        Mostly deep copy of self, with an exception for nodes' attributes, which are shallow-copied, and their `tag` links (no copy).
        """
        dup = copy(self)
        dup.nodes = [node.copy() for node in self.nodes]
        return dup
        

    ### SELECTORS API
    
    def walk(self, skip = None, order = 'preorder'):
        """
        Generator of all nodes inside this DOM: parent nodes and descendants.
        Parents are yielded before descendants if order='preorder' (default), or after descendants if order='postorder'.
        An optional argument `skip` is a function that takes a Node instance and returns True if the subtree
        rooted at this node should be omitted, False otherwise.
        The stream of nodes returned is NOT wrapped in a DOM - use select() instead if you need a DOM.
        """
        # if order not in ('preorder', 'postorder'): raise TypeErrorEx("incorrect value of order ({order})")
        return itertools.chain(*(node.walk(skip, order) for node in self.nodes))

    def alter(self, transform, skip = None, order = 'preorder'):
        """
        Apply `transform` function to nodes in this DOM, and in nested DOMs.
        Inside every DOM instance, starting from `self`, a list of nodes is replaced with a concatenated
        list of nodes returned by transform(child) when applied to each child separately.
        This is done before recursive calls to alter() on child nodes (if order='preorder'),
        or after these calls if order='postorder'.
        The `transform` function can be a generator, or a regular function that returns a list of nodes.
        For each input node, `transform` can return any number of nodes: none, or one, or multiple;
        the input node itself can also be returned.
        The alter() function returns self.
        """
        if order == 'preorder':
            for node in self.nodes: node.alter(transform, skip, order)
        
        new_nodes  = itertools.chain(*(transform(node) for node in self.nodes))
        self.nodes = self._flatten(new_nodes)
        
        if order == 'postorder':
            for node in self.nodes: node.alter(transform, skip, order)

        return self
        

    ATTR_DEFINED = Object(name = 'ATTR_DEFINED')

    def select(self, tag = None, attr = None, value = ATTR_DEFINED, order = 'preorder', **attrs):
        """
        Collect a list of all nodes (including descendants) that match the search criteria and return it wrapped up in a new DOM instance.
        Some of the nodes can be related, i.e., an ancestor and its descendant can be returned together,
        with the ancestor still linking to its children and (directly or indirectly) to this descendant.
        When a non-matching node gets removed, its subtree is NOT removed and descendants can still be included in the result.
        :param tag: desired tag name (<str>) or Tag instance that should be present in a matching node
        """
        tag, name, attrs = self._constraint(tag, attr, value, attrs)
        nodes = [node for node in self.walk(order = order) if self._test(node, tag, name, attrs)]
        
        # nodes = []
        # for node in self.walk(order = order):
        #     if tag  is not None and node.tag is not tag: continue
        #     if name is not None and (node.tag is None or node.tag.name != name): continue
        #
        #     stop = False
        #     for a, v in attrs.items():
        #         if a not in node.kwattrs: stop = True; break
        #         if v is not DOM.ATTR_DEFINED and node.kwattrs[a] != v: stop = True; break
        #     if stop: continue
        #
        #     nodes.append(node)
        
        return DOM(*nodes, _strict = False)
        
        
    def skip(self, tag = None, attr = None, value = ATTR_DEFINED, **attrs):
        """
        Return a copy of `self` that has the same structure (nodes, relations, parents and descendants),
        but the subtrees rooted at nodes matching the provided criteria are removed.
        The original DOM (self) and its nodes are left unmodified. All nodes in the returned DOM are copies
        of nodes in self.
        """
        tag, name, attrs = self._constraint(tag, attr, value, attrs)
        dom = self.copy()
        
        def drop(node):
            """Drop subtrees whose root node satisfies the constraints."""
            if not self._test(node, tag, name, attrs):
                yield node
        
        return dom.alter(drop)


    def _constraint(self, tag, attr, value, attrs):
        """Normalization of constraints for select() and skip()."""
        
        name = None
        if tag:
            if isinstance(tag, str):
                name = tag
                tag  = None
        if attr is not None:
            attrs[attr] = value
            
        return tag, name, attrs
        
    def _test(self, node, tag, name, attrs):
        """Returns True if the `node` satisfies all of the constraints."""
        
        if tag  is not None and node.tag is not tag: return False
        if name is not None and (node.tag is None or node.tag.name != name): return False
        
        for a, v in attrs.items():
            if a not in node.kwattrs: return False
            if v is not DOM.ATTR_DEFINED and node.kwattrs[a] != v: return False

        return True
    
        
    # Selectors TODO:
    # - https://github.com/scrapy/cssselect (Scrapy converts all CSS selectors to XPath)
    

    ####################################################################################################################
    #####
    #####  DOM nodes
    #####
    
    class Node:
        """"""
        
        tag     = None      # Tag instance whose expand() will be called to post-process the body, in a non-terminal node
        attrs   = None      # list of positional (unnamed) attributes to be passed to tag.expand() during rendering
        kwattrs = None      # dict of keyword (named) attributes to be passed to tag.expand()
        
        body    = None      # DOM (possibly empty) containing child nodes of a non-terminal node; None in DOM.Text
        
        outline = False     # True/False denotes an "outline" block or an "inline" node; adds a leading newline during rendering if True
        indent  = None      # indentation string of this block: absolute (when starts with \n) or relative
                            # to its parent (otherwise); None means this is an inline (headline) block, no indentation
    
        def __init__(self, body = None, indent = None, **params):
            
            # assign secondary parameters
            for name, value in params.items():
                setattr(self, name, value)
                
            # assign a list of body nodes, with flattening of nested lists and filtering of None's
            self.body = DOM(body)
            
            # assign indentation, with proper handling of absolute (in parent) vs. relative (in children) indentations
            self.set_indent(indent)
            
            # assert not self.tag or isinstance(self.tag, Tag)
            
        def __getitem__(self, attr):
            """Return value of a keyword attribute `attr` if present. Raise an exception otherwise. Shortcut for self.kwattrs[attr]."""
            return self.kwattrs[attr]
            
        def get(self, attr, default = None):
            """Return value of a keyword attribute `attr` if present, or `default` otherwise. Shortcut for self.kwattrs.get(attr, default)."""
            return self.kwattrs.get(attr, default) if self.kwattrs else default
            
        def set_outline(self, outline = True):
            self.outline = outline
    
        def set_indent(self, indent):
            """
            Sets absolute indentation on self. This calls relative_indent() on all children
            to make their indentations relative to the parent's.
            """
            self.indent = indent
            if self.indent is not None:
                for child in self.body:
                    child.relative_indent(self.indent)
    
        def relative_indent(self, parent_indent):
            """
            Convert self.indent from absolute to relative by subtracting `parent_indent`.
            If this node is inline (indent=None), the method is called recursively on child nodes.
            """
            if self.indent is None:
                for child in self.body: child.relative_indent(parent_indent)
            elif self.indent[:1] == '\n':
                assert self.indent.startswith(parent_indent)
                self.indent = self.indent[len(parent_indent):]
            else:
                pass        # self.indent is relative already
                
        def render(self):
            
            text = self.outline * '\n' + self._render_body()
            
            # if self.outline and self.indent:
            if self.indent is not None:
                assert self.indent[:1] != '\n'                      # self.indent must have been converted already to relative
                text = add_indent(text, self.indent)                # indentation is only added where \n is present, the 1st line is left untouched!
                # text = text.replace('\n', '\n' + self.indent)
                
            return text
        
        def _render_body(self):
            
            if self.tag:
                return self.tag.expand(self.body, self.attrs or (), self.kwattrs or {})
            else:
                return self.body.render()
            
        def tree(self, indent = '', step = '  '):
            r"""Return a multiline \n-terminated string that presents this node's structure as a tree."""
            
            name = self.tag.name if self.tag else "<%s>" % self.__class__.__name__
            heading = str(name)
            
            for attr in self.attrs or []:
                heading += " %s" % attr
            for key, attr in (self.kwattrs or {}).items():
                heading += " %s=%s" % (key, attr)
                
            return indent + heading + '\n' + self.body.tree(indent + step, step)
        
        def copy(self):
            """Deep copy of self, with an exception for attributes, which are shallow-copied, and self.tag (no copy)."""
            
            dup = copy(self)
            dup.attrs = copy(self.attrs)
            dup.kwattrs = copy(self.kwattrs)
            
            if self.body is not None:
                dup.body = self.body.copy()
            
            return dup
            
        
        def walk(self, skip = None, order = 'preorder'):
            """
            Generator of all nodes inside the tree rooted at self: parent nodes and descendants.
            Parents are yielded before descendants if order='preorder' (default), or after descendants if order='postorder'.
            An optional argument `skip` is a function that takes a Node instance and returns True if the subtree
            rooted at this node should be omitted, False otherwise.
            """
            if skip is not None and skip(self): return
            if order == 'preorder': yield self
            if self.body:
                for node in self.body.walk(skip, order): yield node
            if order == 'postorder': yield self

        def alter(self, transform, skip = None, order = 'preorder'):
            """Transforms a list of child nodes by calling alter() on self.body. See DOM.alter() for details."""
            
            if skip is not None and skip(self): return self
            if self.body: self.body.alter(transform, skip, order)
            return self
            

    class Root(Node):
        """Root node of a Hypertag DOM tree."""
    
        def render(self, drop_line = True):
            """
            If this Root node represents an entire translated document that was originally fed to a Hypertag parser,
            an extra empty line have been prepended by the parser in Grammar.preprocess() and should be removed now
            - set drop_line=True (default) to perform this correction or drop_line=False to skip it.
            """
            output = super(DOM.Root, self).render()
            return output[1:] if drop_line and output.startswith('\n') else output
            # if not output or not drop_line: return output
            # assert output[0] == '\n'
            # return output[1:]
            
    
    class Text(Node):
        """A leaf node containing plain text."""
        
        text = None         # text of this node, either plain text or markup after preprocessing; consists of two parts:
                            #  1) headline (head) - 1st line of `text`, without trailing newline
                            #  2) tailtext (tail) - all lines after the 1st one including the leading newline (!);
                            #     tailtext may contain trailing newline(s), but this is not obligatory
        
        def __init__(self, text = '', **kwargs):
            super(DOM.Text, self).__init__(text = text, **kwargs)
        
        def __str__(self):
            return self.text
            
        def set_indent(self, indent):
            self.indent = indent
    
        def _render_body(self):
            return self.text
    
        # def indent(self, spaces = 1, gap = 0, re_lines = re.compile(r'^(\s*\n)+|\s+$')):
        #     """
        #     Like self.text, but with leading/trailing empty lines removed and indentation fixed at a given number of `spaces`.
        #     Optionally, a fixed number (`gap`) of empty lines are added at the beginning.
        #     """
        #     text = self.text
        #     text = re_lines.sub('', text)           # strip leading and trailing empty lines
        #
        #     # replace current indentation with a `spaces` number of spaces; existing tabs treated like a single space
        #     lines  = text.splitlines()
        #     indent = ' ' * spaces
        #     offset = min(len(line) - len(line.lstrip()) for line in lines if line.strip())
        #     text   = '\n'.join(indent + line[offset:] for line in lines)
        #     return gap * '\n' + text
        
        
